package org.objectweb.petals.jbi.descriptor;

import java.util.List;

/**
 * Component production is used to describe a component installation
 * package.
 *
 * @version	$Rev: 76 $ $Date: 2006-01-17 18:15:44 +0100 (mar., 17 janv. 2006) $
 * @since	Petals 1.0
 * @author	<a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class ComponentDescription {

	/**
	 * Compoment type: <code>"service-engine"</code>
	 */
	private static String SERVICE_ENGINE_TYPE = "service-engine";

	/**
	 * Component type: <code>"binding-component"</code>
	 */
	private static String BINDING_COMPONENT_TYPE = "binding-component";

	/**
	 * Compoment type: SERVICE_ENGINE_TYPE or BINDING_COMPOMENT_TYPE.
	 */
	private String type;

	/**
	 * Component's class loading delegation model: <code>"self-first"</code>
	 * or <code>"parent-first"</code>.
	 */
	private String componentClassLoaderDelegation;

	/**
	 * Component's bootstrap class loading delegation model:
	 * <code>"self-first"</code> or <code>"parent-first"</code>.
	 */
	private String bootstrapClassLoaderDelegation;

	/**
	 * Component identification data.
	 */
	private Identification identification;

	/**
	 * Fully qualified name for component's <code>Component</code>
	 * implementation.
	 */
	private String componentClassName;

	/**
	 * Component Class Path.
	 */
	private List<String> componentClassPath;

	/**
	 * Fully qualified name for bootstrap's <code>Bootstrap</code>
	 * implementation.
	 */
	private String bootstrapClassName;

	/**
	 * Bootstrap Class Path.
	 */
	private List<String> bootstrapClassPath;

	/**
	 * List of Shared Libraries used by this component.
	 */
	private List<SharedLibraryList> sharedLibraryList;

	/**
	 * Default constructor.
	 */
	ComponentDescription() {
		super();
	}

	//
	// Getters and Setters
	//

	/**
	 * Returns the class loading delegation model of component's bootstrap
	 * used during component installation. If <code>null</code> the
	 * "parent-first" default delegation model must be used.
	 *
	 * @return <code>String</code> the delegation model:
	 * <code>"parent-first"</code> or <code>"self-first"</code>.
	 */
	public String getBootstrapClassLoaderDelegation() {
		return bootstrapClassLoaderDelegation;
	}

	/**
	 * Returns the fully-qualified class name for the component's
	 * <code>Bootstrap</code> implementation. This is used during
	 * component installation.
	 *
	 * @return <code>String</code> the fully-qualified class name.
	 */
	public String getBootstrapClassName() {
		return bootstrapClassName;
	}

	/**
	 * Returns the class path used during installation processing
	 * (bootstrapping) of the component.
	 *
	 * @return <code>List<String></code> the bootstrap class path.
	 */
	public List<String> getBootstrapClassPath() {
		return bootstrapClassPath;
	}

	/**
	 * Returns the class loading delegation model of this component used
	 * during its execution phase. If <code>null</code> the "parent-first"
	 * default delegation model must be used.
	 *
	 * @return <code>String</code> the delegation model:
	 * <code>"parent-first"</code> or <code>"self-first"</code>.
	 */
	public String getComponentClassLoaderDelegation() {
		return componentClassLoaderDelegation;
	}

	/**
	 * Returns the fully-qualified class name for the component's
	 * <code>Component</code> implementation. This is used during
	 * component's execution phase of its lifecycle.
	 *
	 * @return <code>String</code> the fully-qualified class name.
	 */
	public String getComponentClassName() {
		return componentClassName;
	}

	/**
	 * Returns the class path used after the component is installed. This
	 * class path, along with shared-libraries, constitutes the full class
	 * path avaiable to the component through its default class loader
	 * context as provided by JBI.
	 *
	 * @return <code>List<String></code> the component class path.
	 */
	public List<String> getComponentClassPath() {
		return componentClassPath;
	}

	/**
	 * Returns component's identification data.
	 *
	 * @return <code>Identification</code> the component's identification.
	 */
	public Identification getIdentification() {
		return identification;
	}

	/**
	 * Returns a list of shared libraries (by name and optional version)
	 * used by the component. The implementation may use the version
	 * information to verify that the shared library version declared by
	 * the component matches the version of the shared library installed
	 * in the JBI environment.
	 *
	 * @return <code>List<SharedLibraryList></code> used by this component.
	 */
	public List<SharedLibraryList> getSharedLibraryList() {
		return sharedLibraryList;
	}

	/**
	 * The component type string representation specifying whether this is
	 * a service engine or a binding component.
	 *
	 * @return <code>String</code> the component type.
	 */
	public String getType() {
		return type;
	}

	/**
	 * Set the Bootstrap class loading delegation model.
	 *
	 * @param delegationModel The delegation model to set:
	 * <code>"parent-first"</code> or <code>"self-first"</code>.
	 */
	void setBootstrapClassLoaderDelegation(String delegationModel) {
		this.bootstrapClassLoaderDelegation = delegationModel;
	}

	/**
	 * Set the class name of component's <code>Bootstrap</code>
	 * implementation. 
	 *
	 * @param className <code>String</code> fully-qualified class name.
	 */
	void setBootstrapClassName(String className) {
		this.bootstrapClassName = className;
	}

	/**
	 * Set the class path used by this component's <code>Bootstrap</code>
	 * during installation process.
	 *
	 * @param classPath <code>List<String></code> the bootstrap class path.
	 */
	void setBootstrapClassPath(List<String> classPath) {
		this.bootstrapClassPath = classPath;
	}

	/**
	 * Set the Component class loading delegation model.
	 *
	 * @param delegationModel the delegation model to set:
	 * <code>"parent-first"</code> or <code>"self-first"</code>.
	 */
	void setComponentClassLoaderDelegation(String delegationModel) {
		this.componentClassLoaderDelegation = delegationModel;
	}

	/**
	 * Set the class name of this component implementation. 
	 *
	 * @param className <code>String</code> fully-qualified class name.
	 */
	void setComponentClassName(String className) {
		this.componentClassName = className;
	}

	/**
	 * Set the class path used by this component during its execution
	 * processing.
	 *
	 * @param classPath <code>List<String></code> the component class path.
	 */
	void setComponentClassPath(List<String> classPath) {
		this.componentClassPath = classPath;
	}

	/**
	 * Set the component identification data.
	 *
	 * @param identification The identification to set.
	 */
	void setIdentification(Identification identification) {
		this.identification = identification;
	}

	/**
	 * Set shared library list used by this component.
	 *
	 * @param sharedLibraryList <code>List<SharedLibraryList></code> shared
	 * library list to set.
	 */
	void setSharedLibraryList(
			List<SharedLibraryList> sharedLibraryList) {

		this.sharedLibraryList = sharedLibraryList;
	}

	/**
	 * Set the component type: <code>SERVICE_ENGINE_TYPE</code> or
	 * <code>BINDING_COMPONENT_TYPE</code>.
	 *
	 * @param type <code>String</code> the component type to set.
	 */
	void setType(String type) {
		this.type = type;
	}

	//
	// Implementation methods
	//

	/**
	 * Returns <code>true</code> if this component type is a Service Engine.
	 * Otherwise returns <code>false</code>.
	 */
	public boolean isServiceEngine() {
		return SERVICE_ENGINE_TYPE.equals(type);
	}

	/**
	 * Returns <code>true</code> if this component type is a Binding
	 * Component. Otherwise returns <code>false</code>.
	 */
	public boolean isBindingComponent() {
		return BINDING_COMPONENT_TYPE.equals(type);
	}
}