package org.objectweb.petals.jbi.descriptor;

import javax.xml.namespace.QName;

/**
 * The connection element declares a mapping from consumer-provided service
 * address to a provider's service endpoint address. This mapping applies
 * to consumption of services by all components. The consumer subelement is
 * used to declare the type of consumer address that is to be mapped (or
 * <i>connected</i>). This declaration can take two forms:
 * <ul>
 * <li><b>service type.</b> This specifies a service type, by supplying an
 * interface-name attribute. All applicable exchanges that are addressed
 * to the given service type are to be sent to the given service provider
 * endpoint instead.</li>
 * <li><b>service endpoint.</b> This specifies a service endpoint, by
 * supplying a service-name attribute and an endpoint-name attribute. All
 * applicable exchanges that are addressed to the given service endpoint
 * are to be sent to the given service provider endpoint instead.</li>
 * </ul>
 * The provider attributes is used to declare the service provider endpoint
 * to which the consumer is to be connected.
 *
 * @version	$Rev: 76 $ $Date: 2006-01-17 18:15:44 +0100 (mar., 17 janv. 2006) $
 * @since	Petals 1.0
 * @author	<a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class Connection {

	/**
	 * Consumer Interface Name for a service type connection.
	 */
	private QName consumerInterfaceName;

	/**
	 * Consumer Service Name for a service endpoint connection.
	 */
	private QName consumerServiceName;

	/**
	 * Consumer Endpoint Name for a service endpoint connection.
	 */
	private String consumerEndpointName;

	/**
	 * Provider Service Name to be mapped the connection.
	 */
	private QName providerServiceName;

	/**
	 * Provider Endpoint Name to be mapped the connection.
	 */
	private String providerEndpointName;

	/**
	 * Default constructor.
	 */
	Connection() {
		super();
	}

	//
	// Getters and Setters
	//

	/**
	 * Returns the consumer Endpoint Name for a <b>service endpoint</b>
	 * connection.
	 *
	 * @return <code>String</code> the endpoint name.
	 */
	public String getConsumerEndpointName() {
		return consumerEndpointName;
	}

	/**
	 * Returns the consumer Interface Name for a <b>service type</b>
	 * connection.
	 *
	 * @return <code>QName</code> the interface name.
	 */
	public QName getConsumerInterfaceName() {
		return consumerInterfaceName;
	}

	/**
	 * Returns the consumer Service Name for a <b>service endpoint</b>
	 * connection.
	 *
	 * @return <code>QName</code> the service name.
	 */
	public QName getConsumerServiceName() {
		return consumerServiceName;
	}

	/**
	 * Returns the provider Endpoint Name which the connection is to be
	 * mapped.
	 *
	 * @return <code>String</code> the endpoint name.
	 */
	public String getProviderEndpointName() {
		return providerEndpointName;
	}

	/**
	 * Returns the provider Service Name which the connection is to be
	 * mapped.
	 *
	 * @return <code>QName</code> the service name.
	 */
	public QName getProviderServiceName() {
		return providerServiceName;
	}

	/**
	 * Set the consumer Endpoint Name for a <b>service endpoint</b>
	 * connection.
	 *
	 * @param endpointName the endpoint name.
	 */
	void setConsumerEndpointName(String endpointName) {
		this.consumerEndpointName = endpointName;
	}

	/**
	 * Set the consumer Interface Name for a <b>service type</b> connection.
	 *
	 * @param interfaceName the interface name.
	 */
	void setConsumerInterfaceName(QName interfaceName) {
		this.consumerInterfaceName = interfaceName;
	}

	/**
	 * Set the consumer Service Name for a <b>service endpoint</b>
	 * connection.
	 *
	 * @param serviceName the service name.
	 */
	void setConsumerServiceName(QName serviceName) {
		this.consumerServiceName = serviceName;
	}

	/**
	 * Set the provider Endpoint Name to be mapped.
	 *
	 * @param endpointName the endpoint name.
	 */
	void setProviderEndpointName(String endpointName) {
		this.providerEndpointName = endpointName;
	}

	/**
	 * Set the provder Service Name to be mapped.
	 *
	 * @param serviceName the service name.
	 */
	void setProviderServiceName(QName serviceName) {
		this.providerServiceName = serviceName;
	}
	
}