/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIDescriptorBuilder.java 216 2006-04-13 14:18:14Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.descriptor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.util.XmlHelper;
import org.objectweb.petals.util.FileUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * TODO refactor algorithm
 * The JBIDescriptorBuilder helps to populate a <code>JBIDescriptor</code>
 * instance with values loaded from a specified <i>jbi.xml</i>.
 *
 * @version	$Rev: 216 $ $Date: 2006-04-13 16:18:14 +0200 (jeu., 13 avr. 2006) $
 * @since	Petals 1.0
 * @author	<a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class JBIDescriptorBuilder {

	/**
	 * Standard location to the JBI Descriptor resource.
	 */
	public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF/jbi.xml";

	/**
	 * The JBI Descriptor XML Schema Resource Location ("schema/jbi.xsd").
	 */
	public static final String JBI_SCHEMA_RESOURCE = "schema/jbi.xsd";

	/**
	 * Logger wrapper
	 */
	private LoggingUtil log;
	
	/**
	 * This class is used to handle errors during creating XML objects,
	 * parsing XML and validating a XML.
	 */
	private static class JbiXmlErrorHandler implements ErrorHandler {

		private LoggingUtil log;
		
		/**
		 *
		 */
		public JbiXmlErrorHandler(LoggingUtil loggingUtil) 
		{
			super();
			log = loggingUtil;
		}

		/**
		 * @see org.xml.sax.ErrorHandler#warning(org.xml.sax.SAXParseException)
		 */
		public void warning(SAXParseException e) throws SAXException {
			log.warning(e.getMessage(),e);
		}

		/**
		 * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
		 */
		public void error(SAXParseException se) throws SAXException {
			log.error(se.getMessage(),se);
			throw se;
		}

		/**
		 * @see org.xml.sax.ErrorHandler#fatalError(org.xml.sax.SAXParseException)
		 */
		public void fatalError(SAXParseException se) throws SAXException {
			log.error(se.getMessage(),se);
			throw se;
		}

	}

	/**
	 * A file instance of the JBI Descriptor XML Schema resource. 
	 */
	private static File theJbiSchemaResource;

	/**
	 * URI to the jbi.xml file.
	 */
	private URI jbiXmlUri;

	/**
	 * The source XML document file.
	 */
	private Document source;

	/**
	 * JBI Descriptor Builder constructor using the source xml file URI from
	 * which the document will be loaded.
	 */
	public JBIDescriptorBuilder(URI sourceUri, Logger logger) 
	{
		if (sourceUri == null) {
			throw new IllegalArgumentException(
					"DescriptorBuilder: invalid source URI argument.");
		}

		this.jbiXmlUri = sourceUri;
		
		log = new LoggingUtil(logger);
		
		log.call();
	}

	// ----------------------------------------------------------------------
	//   Getters and Setters
	// ----------------------------------------------------------------------

	/**
	 * @return Returns the jbiXmlUri.
	 */
	public URI getURI() {
		return jbiXmlUri;
	}

	/**
	 * @return Returns the source.
	 */
	public Document getSource() {
		return source;
	}

	// ----------------------------------------------------------------------
	//   implementation methods
	// ----------------------------------------------------------------------

	/**
	 * Load, Parse and Validate jbi.xml document for building (populating) the
	 * <code>Descriptor</code> object.
	 * @throws PetalsException unexpected exception.
	 * @throws SAXException in case a XML validation error occur.
	 */
	public JBIDescriptor build() throws PetalsException, SAXException 
	{
		log.start();
		
		// do not repeat load, parse and validation steps
		if (source == null) {

			// validate the jbi.xml file against the schema
			try 
			{
				validateJbiXml(jbiXmlUri);

				// parse XML after validation
				source = parseJbiXml(jbiXmlUri);

			} 
			catch (IOException ioe) 
			{
				log.error(ioe.getMessage(),ioe);
			
				throw new PetalsException(
						"Unexpected I/O exception using resource: "
						+ jbiXmlUri.toString(), ioe);
			}
		}
		JBIDescriptor result =  buildDescriptor(source);
		
		log.end();
		
		return result;
	}

	/**
	 * 
	 * @param sourceURI
	 * @return
	 * @throws SAXException
	 * @throws IOException
	 */
	public Document parseJbiXml(URI sourceURI) throws SAXException, IOException 
	{
		log.start();
		
		DocumentBuilderFactory docBuilderFactory =
			DocumentBuilderFactory.newInstance();

        DocumentBuilder docBuilder = null;
		
        try 
		{
			docBuilder = docBuilderFactory.newDocumentBuilder();
		} 
        catch (ParserConfigurationException pce) 
        {
			log.error(pce.getMessage(),pce);
			log.end();
			return null;
		}

        // set up xml parsing error handler
        JbiXmlErrorHandler xmlErrorHandler = new JbiXmlErrorHandler(log);
        docBuilder.setErrorHandler(xmlErrorHandler);

        // parse the xml document URI to return
        String jbiXmlURI = sourceURI.toString();

        Document document = null;

        try 
        {
			document = docBuilder.parse(jbiXmlURI);
	    } 
        catch (SAXException saxe) 
        {
        	log.error(saxe.getMessage(),saxe);
	    } 
        catch (IOException ioe) 
        {
        	log.error(ioe.getMessage(),ioe);
		}
        log.end();
        
	    return document;
	}

	/**
	 * 
	 * @param sourceURI
	 * @return
	 * @throws SAXException
	 * @throws IOException
	 */
	public boolean validateJbiXml(URI sourceURI) throws SAXException, IOException 
	{
		log.start();
		
		if (sourceURI == null) {
			return false;
		}

		File document = new File(sourceURI);

        // instance for xml error handling
        JbiXmlErrorHandler xmlErrorHandler = new JbiXmlErrorHandler(log);

        // create a SchemaFactory capable of understanding WXS schemas
	    SchemaFactory factory = SchemaFactory.newInstance(
	    		XMLConstants.W3C_XML_SCHEMA_NS_URI);

	    factory.setErrorHandler(xmlErrorHandler);

	    // load a WXS schema, represented by a Schema instance
	    Source schemaFile = new StreamSource(openJbiSchemaFile());
	    Schema schema;

		try 
		{
			schema = factory.newSchema(schemaFile);
		} 
		catch (SAXException saxe) 
		{
			log.error(saxe.getMessage(),saxe);
			log.end();
			return false;
		}

	    // create a Validator instance which is be used to validate a document
	    Validator validator = schema.newValidator();
	    validator.setErrorHandler(xmlErrorHandler);

	    // validate the DOM tree
	    try 
	    {
	        validator.validate(new StreamSource(document));
	    } 
	    catch (SAXParseException spe) 
	    {
	    	// error thrown by error handler
	    	log.end();
	    	return false;
	    } 
	    catch (SAXException saxe) 
	    {
	    	log.error(saxe.getMessage(),saxe);
	        throw saxe;
	    } 
	    catch (IOException ioe) 
	    {
	    	log.error(ioe.getMessage(),ioe);
	        throw ioe;
		}
	    log.end();
	    
	    // succesfully validated without errors
	    return true;
	}
	
	/**
	 * Returns a String representation of a JBI descriptor file
	 * @param descriptorFile the JBI Descriptor <code>File</code>
	 * @return a <code>String</code> representation of the JBI Descriptor
	 */
	public String getJbiDescriptorAsString(){
		File descriptorFile = new File(jbiXmlUri);
		if (descriptorFile.isFile()) {
			ByteArrayOutputStream os = null;
			InputStream is = null;
            try {
                os = new ByteArrayOutputStream();
                is = new FileInputStream(descriptorFile);
                FileUtil.copyInputStream(is, os);
                return os.toString();
            } catch (Exception e) {
                log.error("Error reading jbi descritor: " + descriptorFile, e);
            }
        }
        return null;
	}

	/**
	 * NOTE: The first location to search for JBI schema files is within the
	 * Petals installation root directory. If a schema file is not found within
	 * root directory, the alternative is use the packaged schema files flatten
	 * somewhere which can be found in class path. It won't work if the schema
	 * are compressed in a Zip or Jar archive.
	 *
	 * @return
	 */
	private File openJbiSchemaFile() 
	{
		log.start();
		
		if (theJbiSchemaResource == null) {

			// find in petals installation root
			File rootDir = SystemUtil.getPetalsInstallDirectory();
			if (rootDir != null) {
				theJbiSchemaResource = new File(rootDir,
						JBI_SCHEMA_RESOURCE);
			}

			// find through class path
			if (theJbiSchemaResource == null) {
				// obtain class loader
				ClassLoader cl = Thread.currentThread().getContextClassLoader();
		
				if (cl == null) {
					cl = JBIDescriptorBuilder.class.getClassLoader();
				}
		
				// load jbi.xsd schema resource file
				URL schemaURL = cl.getResource(
						JBI_SCHEMA_RESOURCE);
				try 
				{
					theJbiSchemaResource = new File(schemaURL.toURI());
				} 
				catch (URISyntaxException e) 
				{
					log.error(e.getMessage(),e);
				}
			}
		}
		log.end();
		
		return theJbiSchemaResource;

	}

	/**
	 * Parse the XML document and load values to the Descriptor object
	 * representation.
	 * 
	 * @param source
	 * @return a populated Descriptor instance with values loaded from the
	 * XML file. May return <code>null</code> if an error occur while navigating
	 * on XML tree.
	 */
	protected JBIDescriptor buildDescriptor(Document source) 
	{
		log.start();
		
		JBIDescriptor descriptor = new JBIDescriptor();

		Node productionTree;

		try 
		{
			// get root element
			Element jbi = source.getDocumentElement();

			// extract JBI version attribute from the root element
			String versionAttr = jbi.getAttribute("version");
			descriptor.setVersion( Double.parseDouble(versionAttr) );

			// component production
			if ((productionTree = XPathAPI.selectSingleNode(source,
					"jbi/component")) != null) {
				descriptor.setComponent( loadComponentXml(productionTree) );
			}
			// shared library production
			else if ((productionTree = XPathAPI.selectSingleNode(source,
					"jbi/shared-library")) != null) {
				descriptor.setSharedLibrary(
						loadSharedLibraryXml(productionTree) );
			}

			// service assembly production
			else if ((productionTree = XPathAPI.selectSingleNode(source, "jbi/service-assembly")) != null) {
				descriptor.setServiceAssembly(
						loadServiceAssemblyXml(productionTree) );
			}

			// services production
			else if ((productionTree = XPathAPI.selectSingleNode(source, "jbi/services")) != null) {
				descriptor.setServices(
						loadServicesXml(productionTree) );
			}

			// no production found
			else {
				log.end();
				return null;
			}

		} catch (TransformerException te) {
			log.error(te.getMessage(),te);
			log.end();
			return null;
		}
		log.end();
		
		return descriptor;
	}

	// ----------------------------------------------------------------------
	//   private static methods
	// ----------------------------------------------------------------------

	/**
	 * @param element
	 * @throws TransformerException 
	 */
	private ComponentDescription loadComponentXml(Node production)
			throws TransformerException 
	{
		ComponentDescription component = new ComponentDescription();
	
		// component attributes
		component.setType( XmlHelper.getAttribute(production, "type") );
		component.setComponentClassLoaderDelegation( XmlHelper.getAttribute(
				production, "component-class-loader-delegation") );
		component.setBootstrapClassLoaderDelegation( XmlHelper.getAttribute(
				production, "bootstrap-class-loader-delegation") );
	
		// identification
		component.setIdentification( loadIdentificationXml(
				XPathAPI.selectSingleNode(production, "identification")) );
	
		// component class name
		component.setComponentClassName( XmlHelper.getText(
				XPathAPI.selectSingleNode(production, "component-class-name")));
	
		// component class path
		component.setComponentClassPath( XmlHelper.getSimpleArray(
				XPathAPI.selectNodeList(production,
						"component-class-path/path-element")) );
	
		// bootstrap class name
		component.setBootstrapClassName( XmlHelper.getText(
				XPathAPI.selectSingleNode(production, "bootstrap-class-name")));
	
		// bootstrap class path
		component.setBootstrapClassPath( XmlHelper.getSimpleArray(
				XPathAPI.selectNodeList(production,
						"bootstrap-class-path/path-element")) );
	
		// shared library list
		component.setSharedLibraryList( loadSharedLibraryListXml(
				XPathAPI.selectNodeList(production, "shared-library-list")) );
	
		return component;
	}

	/**
	 * @param list
	 * @return
	 */
	private List<SharedLibraryList> loadSharedLibraryListXml(NodeList list) 
	{
		if (list == null) {
			return null;
		}

		int size = list.getLength();

		if (size == 0) {
			return null;
		}

		List<SharedLibraryList> result = new ArrayList<SharedLibraryList>();

		for (int i=0; i<size; i++) {
			Node sllElement = list.item(i);

			// shared library list object
			SharedLibraryList newItem = new SharedLibraryList();

			newItem.setName( XmlHelper.getText(sllElement) );
			newItem.setVersion( XmlHelper.getAttribute(sllElement, "version") );

			result.add( newItem );
		}
		return result;
	}

	/**
	 * @param element
	 * @throws TransformerException 
	 */
	private SharedLibrary loadSharedLibraryXml(Node production) throws TransformerException 
	{
		SharedLibrary sharedLibrary = new SharedLibrary();

		// shared library attributes
		sharedLibrary.setClassLoaderDelegation( XmlHelper.getAttribute(
				production, "class-loader-delegation") );
		sharedLibrary.setVersion( XmlHelper.getAttribute(
				production, "version") );

		// identification
		sharedLibrary.setIdentification( loadIdentificationXml(
				XPathAPI.selectSingleNode(production, "identification")) );

		// shared library class path
		sharedLibrary.setSharedLibraryClassPath( XmlHelper.getSimpleArray(
				XPathAPI.selectNodeList(production,
						"shared-library-class-path/path-element")) );
		return sharedLibrary;
	}

	/**
	 * @param element
	 * @throws TransformerException 
	 */
	private ServiceAssembly loadServiceAssemblyXml(Node production) throws TransformerException 
	{
		ServiceAssembly serviceAssembly = new ServiceAssembly();

		// identification
		serviceAssembly.setIdentification( loadIdentificationXml(
				XPathAPI.selectSingleNode(production, "identification")) );

		// service unit
		serviceAssembly.setServiceUnits( loadServiceUnitListXml (
				XPathAPI.selectNodeList(production, "service-unit")) );

		// connections
		serviceAssembly.setConnections( loadConnectionsXml(
				XPathAPI.selectSingleNode(production, "connections")) );

		return serviceAssembly;
	}

	/**
	 * @param connectionsTree
	 * @return
	 * @throws TransformerException 
	 */
	private List<Connection> loadConnectionsXml(Node connectionsTree) throws TransformerException 
	{
		if (connectionsTree == null) {
			log.end();
			return null;
		}

		// get "connections/connection" Node List
		NodeList list = connectionsTree.getChildNodes();

		int size = list.getLength();

		if (size == 0) {
			return null;
		}

		List<Connection> result = new ArrayList<Connection>();

		for (int i=0; i<size; i++) {
			Node cnElement = list.item(i);

			// connection object
			Connection newItem = new Connection();

			// consumer
			Node consumer = XPathAPI.selectSingleNode(cnElement, "consumer");

			newItem.setConsumerInterfaceName( extractXmlAttributeQName(
					consumer, "interface-name") );

			newItem.setConsumerServiceName( extractXmlAttributeQName(
					consumer, "service-name") );

			newItem.setConsumerEndpointName( XmlHelper.getAttribute(
					consumer, "endpoint-name") );

			// provider
			Node provider = XPathAPI.selectSingleNode(cnElement, "provider");

			newItem.setProviderServiceName( extractXmlAttributeQName(
					provider, "service-name") );

			newItem.setProviderEndpointName( XmlHelper.getAttribute(
					provider, "endpoint-name") );

			result.add( newItem );
		}
		return result;
	}

	/**
	 * @param list
	 * @return
	 * @throws TransformerException 
	 */
	private List<ServiceUnit> loadServiceUnitListXml(NodeList list) throws TransformerException 
	{
		if (list == null) {
			return null;
		}

		int size = list.getLength();

		if (size == 0) {
			return null;
		}

		List<ServiceUnit> result = new ArrayList<ServiceUnit>();

		for (int i=0; i<size; i++) {
			Node suElement = list.item(i);

			// service unit object
			ServiceUnit newItem = new ServiceUnit();

			// identification
			newItem.setIdentification( loadIdentificationXml(
					XPathAPI.selectSingleNode(suElement, "identification")) );

			// target
			newItem.setTargetArtifactsZip( XmlHelper.getText(
					XPathAPI.selectSingleNode(suElement, "target/artifacts-zip")) );

			newItem.setTargetComponentName( XmlHelper.getText(
					XPathAPI.selectSingleNode(suElement, "target/component-name")) );

			result.add( newItem );
		}
		
		return result;
	}

	private Identification loadIdentificationXml(Node identificationTree)
			throws TransformerException 
	{
		Identification identification = new Identification();

		identification.setName( XmlHelper.getText(XPathAPI.selectSingleNode(
				identificationTree, "name")) );
		identification.setDescription( XmlHelper.getText(XPathAPI.selectSingleNode(
				identificationTree, "description")) );

		return identification;
	}

	/**
	 * @param element
	 * @throws TransformerException 
	 */
	private static Services loadServicesXml(Node production) throws TransformerException {

		Services services = new Services();

		// is binding component (relative to service unit)
		String isBindingComponent = XmlHelper.getAttribute(production,
				"binding-component");
		if (isBindingComponent != null) {
			services.setBindingComponent(Boolean.parseBoolean(
					isBindingComponent));
		}

		// provides
		services.setProvides( loadProvidesXml(XPathAPI.selectNodeList(
				production, "provides")) );

		// consumes
		services.setConsumes( loadConsumesXml(XPathAPI.selectNodeList(
				production, "consumes")) );

		return services;
	}

	/**
	 * @param list
	 * @return
	 */
	private static List<Consumes> loadConsumesXml(NodeList list) {

		if (list == null) {
			return null;
		}

		int size = list.getLength();

		if (size == 0) {
			return null;
		}

		List<Consumes> result = new ArrayList<Consumes>();

		for (int i=0; i<size; i++) {
			Node consElement = list.item(i);

			// consumes object
			Consumes newItem = new Consumes();

			newItem.setInterfaceName( extractXmlAttributeQName(consElement,
					"interface-name") );

			newItem.setServiceName( extractXmlAttributeQName(consElement,
					"service-name") );

			newItem.setEndpointName( XmlHelper.getAttribute(consElement,
					"endpoint-name") );

			newItem.setLinkType( XmlHelper.getAttribute(consElement,
					"link-type") );

			result.add( newItem );
		}

		return result;
	}

	/**
	 * @param list
	 * @return
	 */
	private static List<Provides> loadProvidesXml(NodeList list) {

		if (list == null) {
			return null;
		}

		int size = list.getLength();

		if (size == 0) {
			return null;
		}

		List<Provides> result = new ArrayList<Provides>();

		for (int i=0; i<size; i++) {
			Node provElement = list.item(i);

			// provides object
			Provides newItem = new Provides();

			newItem.setInterfaceName( extractXmlAttributeQName(provElement,
					"interface-name") );

			newItem.setServiceName( extractXmlAttributeQName(provElement,
					"service-name") );

			newItem.setEndpointName( XmlHelper.getAttribute(provElement,
					"endpoint-name") );

			result.add( newItem );
		}

		return result;
	}

	/**
	 * @param node
	 * @param attrName
	 * @return
	 */
	private static QName extractXmlAttributeQName(Node node, String attrName) {

		String attr = XmlHelper.getAttribute(node, attrName);

		return QName.valueOf(attr);
	}

}
