/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationService.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management;

import static org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_ADMIN_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_JMXADMIN_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.PLATFORM_REPOSITORY_NAME;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.AdminServiceMBean;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.objectweb.petals.jbi.descriptor.ServiceAssembly;
import org.objectweb.petals.jbi.descriptor.ServiceUnit;
import org.objectweb.petals.jbi.management.packaging.JBIPackage;
import org.objectweb.petals.jbi.management.packaging.JBIPackageBuilder;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;
import org.xml.sax.SAXException;

/**
 * JBI deployment service implementation.
 * @author Olivier Fabre - EBM Websourcing
 *
 */
public class DeploymentService implements DeploymentServiceMBean,
BindingController, LifeCycleController {
	
	private static final String[] CLIENT_INTERFACES = new String[] {
		PLATFORM_REPOSITORY_NAME, JBI_JMXADMIN_NAME, JBI_ADMIN_NAME};
	
	/**
	 * Platform Component :: Repository Service
	 */
	private RepositoryService repositorySrv;
	
	/**
	 * Monolog Logger instance.
	 */
	private Logger logger;
	
	/**
	 * logger wrapper
	 */
	private LoggingUtil log;
	
	/**
	 * JBI Container :: JMX Agent Admin
	 */
	private ManagerService managerService;
	
	
	/**
	 * JBI Container :: Admin Service
	 */
	private AdminServiceMBean adminSrv;
	
	/**
	 * Component lifecycle state.
	 */
	private String fcState;
	
	/**
	 * Loading packages cache.
	 */
	private List<String> packagesLoadedCache;
	
	/**
	 * Default constructor.
	 */
	public DeploymentService() 
	{
		
	}
	
	// ----------------------------------------------------------------------
	//   JBI DeploymentServiceMBean interface implementation
	// ----------------------------------------------------------------------
	
	public boolean canDeployToComponent(String componentName) {
		LifeCycle clc = managerService.getComponentByName(componentName);
        return clc != null && LifeCycle.STARTED.equals(clc.getCurrentState()) 
        		&& clc.getComponent().getServiceUnitManager() != null;
	}
	
	public String deploy(String serviceAssemblyZipUrl) throws Exception {
		
		log.start();
		String msg;
		String result = null;
		
		JBIPackage installPackage = loadInstallPackage(serviceAssemblyZipUrl);
		
		if (installPackage == null) 
		{
			// a failure is found
			log.end();
			throw new Exception("Bad installation package");
		}
		
		// Get the Service Assembly  description
		JBIDescriptor descriptor = descriptor = installPackage.getDescriptor();
		
		ServiceAssembly saDesc = descriptor.getServiceAssembly();
		
		// Get SA unique identifier from descriptor
		// check if it is already loaded by package uuid
		String saName = saDesc.getIdentification().getName();
		
		if (isLoaded(saName)) 
		{
			// log error and return failure
			msg = "Other installation package already exists loaded in " +
			"the service assembly framework: " + saName;
			log.error(msg);
			
			log.end();
			throw new Exception(msg);
		}
		
		// Explode Service Assembly into the installation repository
		URI installationRoot = expandPackageIntoRepository(saName,
				installPackage.getArchiveURI());
		
		
		// Create a new ServiceAssemblyLifeCycle
		ServiceAssemblyLifeCycle saLifeCycle = new ServiceAssemblyLifeCycle(log, saDesc, repositorySrv, installationRoot);
		
		// TODO : verify that all Service units are present in the package 
		// and that all targeted components are installed in the JBI System
		// and check duplicate service units as describe in spec section 6.6
		
		// Explode each service unit in the SA root.
		// After expanding su, it deploies it to the targeted component.
		for (ServiceUnit suDesc : saDesc.getServiceUnits()) {
			// Expand it
			String suName = suDesc.getIdentification().getName();
			String suZipNAme = suDesc.getTargetArtifactsZip();
			String targetedComponent = suDesc.getTargetComponentName();
			
			URI suZipLocation = getSUZipURI("file://"+installationRoot.getPath()+suZipNAme);
			
			URI suRoot = expandSUIntoSA(suName, suZipLocation, saName);
			
			// Deploy it
			LifeCycle clc = managerService.getComponentByName(targetedComponent);
			if(clc!=null){
				ServiceUnitManager sum = clc.getComponent().getServiceUnitManager();
				if (sum!=null){
					sum.deploy(suName, suRoot.getPath());
					// TODO : add the result of deploy method call on su
					// to the management message "result"
					
					// Create and register a new ServiceUnitLifeCycle
					// TODO : verify deployment success before registration
					ServiceUnitLifeCycle suLifeCycle = new ServiceUnitLifeCycle(suDesc, managerService, 
																				suRoot.getPath(),log);
					saLifeCycle.registerSUIntoSA(suName, suLifeCycle);
				}
			}
		}
		// Register ServiceAssemblyLifeCycle into manager service map.
		managerService.registerServiceAssembly(saName, saLifeCycle);
		
		// TODO : returning xml result instead of SA name
		result = saName;
		return result;
	}
	
	public String[] getComponentsForDeployedServiceAssembly(String serviceAssemblyName) throws Exception {
		List<String> result = new ArrayList<String>();
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		if(saLifeCycle!=null){
			List<ServiceUnit> sus = saLifeCycle.getServiceAssembly().getServiceUnits();
			for (ServiceUnit unit : sus) {
				result.add(unit.getTargetComponentName());
			}
		}else{
			throw new ManagementException("Cannot retrieve service " +
					"assembly with the specified name : "+serviceAssemblyName);
		}
		return result.toArray(new String[0]);
	}
	
	public String[] getDeployedServiceAssemblies() throws Exception {
		Set<String> sas = managerService.getSync_serviceAssemblyLifeCycles().keySet();
		return sas.toArray(new String[0]);
	}
	
	public String[] getDeployedServiceAssembliesForComponent(String componentName) throws Exception {
		Set<String> result = new HashSet<String>();
		Map<String, ServiceAssemblyLifeCycle> saLifeCycles = managerService.getSync_serviceAssemblyLifeCycles();
		for (ServiceAssemblyLifeCycle salc : saLifeCycles.values()) {
			List<ServiceUnit> sus = salc.getServiceAssembly().getServiceUnits();
			for (ServiceUnit unit : sus) {
				if(componentName.equals(unit.getTargetComponentName())){
					result.add(salc.getServiceAssembly().getIdentification().getName());
				}
			}
		}
		return result.toArray(new String[0]);
	}
	
	public String[] getDeployedServiceUnitList(String componentName) throws Exception {
		Set<String> result = new HashSet<String>();
		Map<String, ServiceAssemblyLifeCycle> saLifeCycles = managerService.getSync_serviceAssemblyLifeCycles();
		for (ServiceAssemblyLifeCycle salc : saLifeCycles.values()) {
			List<ServiceUnit> sus = salc.getServiceAssembly().getServiceUnits();
			for (ServiceUnit unit : sus) {
				if(componentName.equals(unit.getTargetComponentName())){
					result.add(unit.getIdentification().getName());
				}
			}
		}
		return result.toArray(new String[0]);
	}
	
	public String getServiceAssemblyDescriptor(String serviceAssemblyName) throws Exception {
		String result = "Service Assembly doesn't exist :"+serviceAssemblyName;
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		if(saLifeCycle!=null){
			URI installationRoot = saLifeCycle.getInstallationRoot();
			// Load JBI descriptor as string
			URI jbiXmlUri = installationRoot.resolve(JBI_DESCRIPTOR_RESOURCE);
			JBIDescriptorBuilder builder = new JBIDescriptorBuilder(jbiXmlUri,logger);
			result = builder.getJbiDescriptorAsString();
		}

		return result;
	}
	
	public String getState(String serviceAssemblyName) throws Exception {
		String result = null;
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		if(saLifeCycle!=null){
			result = saLifeCycle.getCurrentState();
		}else{
			throw new ManagementException("Cannot retrieve service " +
					"assembly with the specified name : "+serviceAssemblyName);
		}
		return result;
	}
	
	public boolean isDeployedServiceUnit(String componentName, String serviceUnitName) throws Exception {
		boolean result = false;
		Map<String, ServiceAssemblyLifeCycle> saLifeCycles = managerService.getSync_serviceAssemblyLifeCycles();
		for (Iterator iter = saLifeCycles.values().iterator(); iter.hasNext() && !result;) {
			ServiceAssemblyLifeCycle salc = (ServiceAssemblyLifeCycle)iter.next();
			List<ServiceUnit> sus = salc.getServiceAssembly().getServiceUnits();
			for (Iterator iter2 = sus.iterator();iter2.hasNext() && !result;) {
				ServiceUnit unit = (ServiceUnit)iter2.next();
				if(componentName.equals(unit.getTargetComponentName())){
					result=true;
				}
			}
		}
		return result;
	}
	
	public String shutDown(String serviceAssemblyName) throws Exception {
		
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		
		if(saLifeCycle!=null){
			saLifeCycle.shutDown();
		}
		
		// TODO : returning xml result instead of SA name
		return serviceAssemblyName;
	}
	
	public String start(String serviceAssemblyName) throws Exception {
		
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		
		if(saLifeCycle!=null){
			saLifeCycle.start();
		}
		
		// TODO : returning xml result instead of SA name
		return serviceAssemblyName;
	}
	
	public String stop(String serviceAssemblyName) throws Exception {
		
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		
		if(saLifeCycle!=null){
			saLifeCycle.stop();
		}	
		
		// TODO : returning xml result instead of SA name
		return serviceAssemblyName;
	}
	
	public String undeploy(String serviceAssemblyName) throws Exception {
		ServiceAssemblyLifeCycle saLifeCycle = managerService.getServiceAssemblyByName(serviceAssemblyName);
		
		if(saLifeCycle!=null){
			saLifeCycle.undeploy();
		}
		
		// Remove the service assembly life cycle from the manager service SA Map
		managerService.unregisterServiceAssembly(serviceAssemblyName);
		
		// TODO : returning xml result instead of SA name
		return serviceAssemblyName;
	}
	
	// ----------------------------------------------------------------------
	//   implementation specific methods
	// ----------------------------------------------------------------------
	
	/**
	 * Load JBI Descriptor from the installation root repository
	 */
	public JBIDescriptor loadInstallationDescriptor(URI installationRoot) 
	{
		log.start();
		
		String msg;
		JBIDescriptor descriptor = null;
		
		try 
		{
			URI jbiXmlUri = installationRoot.resolve(JBI_DESCRIPTOR_RESOURCE);
			
			JBIDescriptorBuilder builder = new JBIDescriptorBuilder(jbiXmlUri,logger);
			
			descriptor = builder.build();
		}
		catch (PetalsException pe) 
		{
			// just log and return failure status
			msg = "Unexpected exception. Failed to load installation root: " +
			installationRoot.toString() + ". " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		} 
		catch (SAXException e) 
		{
			// just log and return failure status
			msg = "Error processing JBI Descriptor XML. Installation root: " +
			installationRoot.toString() + ". " + e.getMessage();
			log.error(msg, e);
			log.end();
			return null;
		}
		log.end();
		
		return descriptor;
	}
	
	
	private URI expandPackageIntoRepository(String uuid, URI zipArchiveURI) 
	{
		log.start();
		
		String msg;
		URI installationRoot = null;
		
		// explode the installation package Zip archive into platform repository
		
		try 
		{
			installationRoot = repositorySrv.addPackage(uuid, zipArchiveURI);
		}
		catch (PetalsException pe) 
		{
			// log error and return failure
			msg = "Unexpected error adding shared library JBI installation " +
			"package into repository. " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		}
		
		if (installationRoot == null) 
		{
			msg = "Unexpected Platform repository service error. The " +
			"installation root must NOT be null: " + uuid;
			log.error(msg);
			
			// don't keep incomplete package installation into repository 
			repositorySrv.removePackage(uuid);
			
			// return failure
			log.end();
			return null;
		}
		log.end();
		
		return installationRoot;
	}
	
	private URI expandSUIntoSA(String suId, 
			URI suZipLocation, 
			String saId)
	{
		log.start();
		
		String msg;
		URI suRoot = null;
		
		// explode the installation package Zip archive into platform repository
		try 
		{
			suRoot = repositorySrv.explodeSUIntoSAInstallFolder(suId, 
					suZipLocation, 
					saId);
		}
		catch (PetalsException pe) 
		{
			// log error and return failure
			msg = "Unexpected error adding shared library JBI installation " +
			"package into repository. " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		}
		
		if (suRoot == null) 
		{
			msg = "Unexpected Platform repository service error. The " +
			"su root must NOT be null: " + suId;
			log.error(msg);
			
			// return failure
			log.end();
			return null;
		}
		log.end();
		
		return suRoot;
	}
	
	
	public boolean isLoaded(String packageUUID) 
	{
		return packagesLoadedCache.contains(packageUUID);
	}
	
	
	/**
	 * @param slZipURL
	 * @return May return <code>null</code> in case of a failure.
	 */
	private JBIPackage loadInstallPackage(String slZipURL) 
	{
		log.start();
		
		String msg;
		JBIPackage jbiPackage;
		
		try 
		{
			JBIPackageBuilder pkgBuilder = new JBIPackageBuilder(slZipURL,logger);
			
			// here happens the XML Schema validation
			jbiPackage = pkgBuilder.build();
		}
		catch (PetalsException pe) 
		{
			// just log and return failure status
			msg = "Failed to open or build the JBI Package: " + slZipURL + 
			". " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		}
		LoggingUtil.end(logger);
		
		return jbiPackage;
	}
	
	
	protected void init() 
	{
		log = new LoggingUtil(logger);
		log.start();
		
		// initialize packages loading cache
		packagesLoadedCache = new ArrayList<String>();
		
		log.end();
	}
	
	protected void shutdown() 
	{
		log.start();
		
		logger = null;
		
		packagesLoadedCache = null;
		
		log.end();
	}
	
	/**
	 * Process the given su Zip archive URL string and transform into a
	 * ready-to-use URI object.
	 * @param suZipURL <code>String</code> the su zip archive URL string
	 * @return a ready-to-use <code>URI</code> object
	 * @throws PetalsException
	 */
	private URI getSUZipURI(String suZipURL) throws PetalsException{
		log.start();
		String msg;
		URI uri = null;
		
		try 
		{
			URL url = new URL(suZipURL);
			
			uri = url.toURI();
		}
		catch (MalformedURLException mue) 
		{
			msg = "Wrong su archive location specified. URL: "
				+ suZipURL;
			
			log.error(msg, mue);
			
			throw new PetalsException(msg, mue);
		}
		catch (URISyntaxException use) 
		{
			msg = "Syntax error converting to URI from given location. URL: "
				+ suZipURL;
			log.error(msg, use);
			
			throw new PetalsException(msg, use);
		}
		log.end();
		return uri;
	}
	
	// ----------------------------------------------------------------------
	//   Fractal BindingController implementation
	// ----------------------------------------------------------------------
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#listFc()
	 */
	public String[] listFc() {
		return CLIENT_INTERFACES;
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 */
	public Object lookupFc(String clientItfName)
	throws NoSuchInterfaceException {
		
		if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			return repositorySrv;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			return managerService;
		}
		else if (JBI_ADMIN_NAME.equals(clientItfName)) {
			return adminSrv;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String, java.lang.Object)
	 */
	public void bindFc(String clientItfName, Object serverItf)
	throws NoSuchInterfaceException, IllegalBindingException,
	IllegalLifeCycleException {
		
		if (LOGGER_NAME.equals(clientItfName)) {
			logger = (Logger) serverItf;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}
		else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			repositorySrv = (RepositoryService) serverItf;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			managerService = (ManagerService) serverItf;
		}
		else if (JBI_ADMIN_NAME.equals(clientItfName)) {
			adminSrv = (AdminServiceMBean) serverItf;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
	IllegalBindingException, IllegalLifeCycleException {
		
		if (LOGGER_NAME.equals(clientItfName)) {
			logger = null;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}
		else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			repositorySrv = null;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			managerService = null;
		}
		else if (JBI_ADMIN_NAME.equals(clientItfName)) {
			adminSrv = null;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	// ----------------------------------------------------------------------
	//   Fractal LifecycleController implementation
	// ----------------------------------------------------------------------
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return (org.objectweb.fractal.api.control.LifeCycleController.STARTED.equals(fcState) ? 
				org.objectweb.fractal.api.control.LifeCycleController.STARTED : 
					org.objectweb.fractal.api.control.LifeCycleController.STOPPED);
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {
		
		init();
		
		fcState = org.objectweb.fractal.api.control.LifeCycleController.STARTED;
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {
		
		shutdown();
		
		fcState = org.objectweb.fractal.api.control.LifeCycleController.STOPPED;
	}
	
}

