/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationService.java 154 2006-03-27 15:30:10Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management;

import static org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_JMXADMIN_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_MESSAGING_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.PLATFORM_CLASSLOADING_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.PLATFORM_REPOSITORY_NAME;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.classloader.LoaderManager;
import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.descriptor.ComponentDescription;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.objectweb.petals.jbi.descriptor.SharedLibrary;
import org.objectweb.petals.jbi.management.autoload.AutoLoaderImpl;
import org.objectweb.petals.jbi.management.packaging.JBIPackage;
import org.objectweb.petals.jbi.management.packaging.JBIPackageBuilder;
import org.objectweb.petals.jbi.messaging.MessagingService;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;
import org.xml.sax.SAXException;

/**
 * JBI installation service implementation.
 *
 * @version	$Rev: 154 $ $Date: 2006-03-27 17:30:10 +0200 (lun., 27 mars 2006) $
 * @since	Petals 1.0
 * @author	<a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class InstallationService implements InstallationServiceMBean,
BindingController, LifeCycleController {
	
	private static final String[] CLIENT_INTERFACES = new String[] {
		PLATFORM_CLASSLOADING_NAME,
		PLATFORM_REPOSITORY_NAME, JBI_JMXADMIN_NAME, JBI_MESSAGING_NAME };
	
	/**
	 * Platform Component :: Loader Service
	 */
	private LoaderManager loaderSrv;
	
	/**
	 * Platform Component :: Repository Service
	 */
	private RepositoryService repositorySrv;
	
	/**
	 * Monolog Logger instance.
	 */
	private Logger logger;
	
	/**
	 * logger wrapper
	 */
	private LoggingUtil log;
	
	/**
	 * JBI Container :: JMX Agent Admin
	 */
	private ManagerService jmxAdmin;
	
	/**
	 * JBI Container :: Messaging Service
	 */
	private MessagingService messagingSrv;
	
	/**
	 * Automatic Loading
	 */
	private AutoLoaderImpl autoLoader;
	
	/**
	 * Component lifecycle state.
	 */
	private String fcState;
	
	/**
	 * Loading packages cache.
	 */
	private List<String> packagesLoadedCache;
	
	/**
	 * Default constructor.
	 */
	public InstallationService() 
	{
		
	}
	
	// ----------------------------------------------------------------------
	//   JBI InstallationServiceMBean interface implementation
	// ----------------------------------------------------------------------
	
	/**
	 * @see javax.jbi.management.InstallationServiceMBean#installSharedLibrary(java.lang.String)
	 * TODO refactor
	 */
	public synchronized String installSharedLibrary(String slZipURL) {
		
		log.start();
		
		String msg;
		String result = null;
		
		/*
		 * Get the right package location for preparing installation.
		 * Resolve, parse and validate the JBI Descriptor from the
		 * package archive Zip file. Thus, load JBI Descriptor with data
		 * from XML for using during this process.
		 */
		
		JBIPackage installPackage = loadInstallPackage(slZipURL);
		
		if (installPackage == null) {
			// a failure is found
			return null;
		}
		
		// Get the shared library production element
		JBIDescriptor descriptor = descriptor = installPackage.getDescriptor();
		SharedLibrary sl = descriptor.getSharedLibrary();
		
		// check if jbi descriptor has no shared-library element
		if (sl == null) {
			msg = "The shared-library production element was not found in " +
			"JBI Descriptor within installation package: " + slZipURL;
			log.error(msg);
			
			return null;
		}
		
		// get shared library unique identifier from descriptor
		
		String slName = sl.getIdentification().getName();
		
		
		// check if it is already loaded by package uuid
		
		if (isLoaded(slName)) 
		{
			// log error and return failure
			msg = "Other installation package already exists loaded in " +
			"the component framework: " + slName;
			
			log.error(msg);
			
			return null;
		}
		
		
		// explode shared library into the installation repository
		
		URI installationRoot = expandPackageIntoRepository(slName,
				installPackage.getArchiveURI());
		
		
		// proceed shared library installation using PetalsInstallationService
		// interface method.
		
		result = installSharedLibrary(slName, installationRoot, descriptor);
		
		log.end();
		
		return result;
	}
	
	/**
	 * @see javax.jbi.management.InstallationServiceMBean#loadInstaller(java.lang.String)
	 */
	public synchronized ObjectName loadInstaller(String componentName) {
		return jmxAdmin.getInstallerByName(componentName);
	}
	
	/**
	 * TODO refactor the algorithm
	 * @see javax.jbi.management.InstallationServiceMBean#loadNewInstaller(java.lang.String)
	 */
	public synchronized ObjectName loadNewInstaller(String installZipURL) 
	{
		log.start();
		
		String msg;
		
		ObjectName installerMBeanName = null;
		
		/*
		 * Get the right package location for preparing installation.
		 * Resolve, parse and validate the JBI Descriptor from the
		 * package archive Zip file. Thus, load JBI Descriptor with data
		 * from XML for using during this process.
		 */
		
		try
		{
			JBIPackage installPackage = loadInstallPackage(installZipURL);
			
			if (installPackage == null) 
			{
				// a failure is found
				log.end();
				return null;
			}
			
			// 1. Get the component production element
			
			JBIDescriptor descriptor = descriptor = installPackage.getDescriptor();
			
			ComponentDescription component = descriptor.getComponent();
			
			
			// 2. check if jbi descriptor has no shared-library element
			
			if (component == null) 
			{
				msg = "The component production element was not found in " +
				"JBI Descriptor within installation package: " +
				installZipURL;
				log.error(msg);
				
				log.end();
				return null;
			}
			
			// 3. get component unique identifier from descriptor
			// check if it is already loaded by package uuid
			
			String componentName = component.getIdentification().getName();
			
			if (isLoaded(componentName)) 
			{
				// log error and return failure
				msg = "Other installation package already exists loaded in " +
				"the component framework: " + componentName;
				log.error(msg);
				
				log.end();
				return null;
			}
			
			// 4. explode component into the installation repository
			
			URI installationRoot = expandPackageIntoRepository(componentName,
					installPackage.getArchiveURI());
			
			
			// 5. prepare the component context
			// JBI spec requires that the installPath mus be in formatted  for the
			// underlying platform
			String installationRootPath = new File(installationRoot).getAbsolutePath();
			String workingRootPath = repositorySrv.getComponentWorkRoot(componentName).getAbsolutePath();
			
			ComponentContextImpl componentContext = new ComponentContextImpl(descriptor, installationRootPath,workingRootPath, messagingSrv, jmxAdmin);
			
			// 6. create and init installer mbean
			try 
			{
				Installer installerMBean = new Installer(
						installationRoot, 
						loaderSrv, 
						componentContext,
						descriptor.getComponent(), 
						jmxAdmin,
						logger);
			
				installerMBeanName = jmxAdmin.registerInstaller(installerMBean);
			} 
			catch (JBIException e) 
			{
				// just log when some error occur
				log.error( 
						"Error creating installerMBean. "
						+ e.getMessage(), e);
				
				log.end();
				return null;
			} 
		} 
		catch (Throwable t)
		{
			log.error(t);
		}
		log.end();
		
		return installerMBeanName;
	}
	
	/**
	 * @see javax.jbi.management.InstallationServiceMBean#uninstallSharedLibrary(java.lang.String)
	 */
	public synchronized boolean uninstallSharedLibrary(String slName) 
	{
		log.start();
		
		boolean result = true;
		
		// unload shared library class loaders and clean cache
		loaderSrv.deleteClassLoader(slName);
		
		packagesLoadedCache.remove(slName);
		
		// clean package installation from platform repository
		result = repositorySrv.removePackage(slName);
		
		log.end();
		
		return result;
	}
	
	/**
	 * TODO isToBeDeleted parameter isn't used for the moment : 
	 * this method completly deletes the component from the container
	 * TODO Problem with jar deletion, they can't be 
	 * deleted before classloader been garbaged collected
	 * @see javax.jbi.management.InstallationServiceMBean#unloadInstaller(java.lang.String, boolean)
	 */
	public synchronized boolean unloadInstaller(String componentName,
			boolean isToBeDeleted) {
		log.start();
		boolean result = false;
		// Call uninstall on Installer
		ObjectName instName = jmxAdmin.getInstallerByName(componentName);
		MBeanServer mbeanSrv = jmxAdmin.getMBeanServer();
		Object[] nullObjects= new Object[0];
		String[] nullStrings= new String[0];
		try {
			mbeanSrv.invoke(instName,"uninstall",nullObjects,nullStrings);
		} catch (InstanceNotFoundException e) {
			// just log when some error occur
			log.error( 
					"Error uninstalling component. "
					+ e.getMessage(), e);
			
			log.end();
			return false;
		} catch (MBeanException e) {
			// just log when some error occur
			log.error( 
					"Error uninstalling component. "
					+ e.getMessage(), e);
			
			log.end();
			return false;
		} catch (ReflectionException e) {
			// just log when some error occur
			log.error( 
					"Error uninstalling component. "
					+ e.getMessage(), e);
			
			log.end();
			return false;
		}
		
		// unload component and bootstrap class loaders
		loaderSrv.deleteClassLoader(Installer.compPrefix+componentName);
		loaderSrv.deleteClassLoader(Installer.bsPrefix+componentName);
		
		// unload installer from manager service
		try {
			jmxAdmin.unregisterInstaller(instName);
		} catch (ManagementException e) {
			// just log when some error occur
			log.error( 
					"Error uninstalling component. "
					+ e.getMessage(), e);
			
			log.end();
			return false;
		}
		
		// clean package installation from platform repository
		// TODO Problem with jar deletion, they can't be 
		// deleted before classloader been garbaged collected
		// result = repositorySrv.removePackage(componentName);
		result = true;
		
		log.end();
		return result;
	}
	
	// ----------------------------------------------------------------------
	//   PetalsInstallationService interface implementation methods
	// ----------------------------------------------------------------------
	
	/**
	 * @see org.objectweb.petals.jbi.management.service.PetalsInstallationService#installSharedLibrary(java.lang.String, java.net.URI, org.objectweb.petals.jbi.descriptor.JBIDescriptor)
	 */
	public synchronized String installSharedLibrary(String slName,
			URI installationRoot, JBIDescriptor descriptor) 
	{
		log.start();
		
		String msg;
		
		// create class loaders and store into platform
		// List<URL> classPathUrls = new ArrayList<URL>();
		SharedLibrary sl = descriptor.getSharedLibrary();
		
		/*for (String pathElement : sl.getSharedLibraryClassPath()) {
		 URL pathElementUrl;
		 
		 try {
		 pathElementUrl = installationRoot.resolve(pathElement).toURL();
		 classPathUrls.add(pathElementUrl);
		 }
		 catch (MalformedURLException e) {
		 // simple do nothing - JBI Package has responsibility in
		  // checking the archive contents against informations
		   // contained in JBI Descriptor first.
		    msg = "Unexpected error. " + e.getMessage();
		    log(BasicLevel.LEVEL_INFO, msg, e);
		    }
		    }*/
		
		/*
		 * Create shared library class loader using 
		 * Use platform service to create and manage class loader
		 */
		
		boolean parentFirst = JBIDescriptor.isParentFirst(
				sl.getClassLoaderDelegation());
		
		//URL[] urls = classPathUrls.toArray(new URL[] { });
		
		List<URL> urls = new ArrayList<URL>();
		
		try 
		{
			urls.add(installationRoot.toURL());
		} 
		catch (MalformedURLException e) 
		{
			// simple do nothing - JBI Package has responsibility in
			// checking the archive contents against informations
			// contained in JBI Descriptor first.
			msg = "Unexpected error. " + e.getMessage();
			log.info(msg, e);
		}
		
		URL[] baseUrls = urls.toArray(new URL[] { });
		
		try 
		{
			loaderSrv.createClassLoader(slName, baseUrls, sl.getSharedLibraryClassPath(), parentFirst);
		}
		catch (PetalsException pe) 
		{
			// just log when some error occur
			log.error(
					"Unexpected error creating class loader. "
					+ pe.getMessage(), pe);
			
			log.end();
			return null;
		}
		
		// add shared library to the loaded cache list
		packagesLoadedCache.add(slName);
		
		log.end();
		
		return slName;
	}
	
	// ----------------------------------------------------------------------
	//   implementation specific methods
	// ----------------------------------------------------------------------
	
	/**
	 * Load JBI Descriptor from the installation root repository
	 */
	public JBIDescriptor loadInstallationDescriptor(URI installationRoot) 
	{
		log.start();
		
		String msg;
		JBIDescriptor descriptor = null;
		
		try 
		{
			URI jbiXmlUri = installationRoot.resolve(JBI_DESCRIPTOR_RESOURCE);
		
			JBIDescriptorBuilder builder = new JBIDescriptorBuilder(jbiXmlUri,logger);
			
			descriptor = builder.build();
		}
		catch (PetalsException pe) 
		{
			// just log and return failure status
			msg = "Unexpected exception. Failed to load installation root: " +
			installationRoot.toString() + ". " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		} 
		catch (SAXException e) 
		{
			// just log and return failure status
			msg = "Error processing JBI Descriptor XML. Installation root: " +
			installationRoot.toString() + ". " + e.getMessage();
			log.error(msg, e);
			log.end();
			return null;
		}
		log.end();
		
		return descriptor;
	}
	
	
	private URI expandPackageIntoRepository(String uuid, URI zipArchiveURI) 
	{
		log.start();
		
		String msg;
		URI installationRoot = null;
		
		// explode the installation package Zip archive into platform repository
		
		try 
		{
			installationRoot = repositorySrv.addPackage(uuid, zipArchiveURI);
		}
		catch (PetalsException pe) 
		{
			// log error and return failure
			msg = "Unexpected error adding shared library JBI installation " +
			"package into repository. " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		}
		
		if (installationRoot == null) 
		{
			msg = "Unexpected Platform repository service error. The " +
			"installation root must NOT be null: " + uuid;
			log.error(msg);
			
			// don't keep incomplete package installation into repository 
			repositorySrv.removePackage(uuid);
			
			// return failure
			log.end();
			return null;
		}
		log.end();
		
		return installationRoot;
	}
	
	
	public boolean isLoaded(String packageUUID) 
	{
		return packagesLoadedCache.contains(packageUUID);
	}
	
	
	/**
	 * @param slZipURL
	 * @return May return <code>null</code> in case of a failure.
	 */
	private JBIPackage loadInstallPackage(String slZipURL) 
	{
		log.start();
			
		String msg;
		JBIPackage jbiPackage;
		
		try 
		{
			JBIPackageBuilder pkgBuilder = new JBIPackageBuilder(slZipURL,logger);
			
			// here happens the XML Schema validation
			jbiPackage = pkgBuilder.build();
		}
		catch (PetalsException pe) 
		{
			// just log and return failure status
			msg = "Failed to open or build the JBI Package: " + slZipURL + 
			". " + pe.getMessage();
			log.error(msg, pe);
			log.end();
			return null;
		}
		
		// shouldn't it be an installation package?
		if (jbiPackage.isDeploymentPackage()) 
		{
			// log and return failure
			msg = "An invalid JBI deployment package was given instead of a " +
			"JBI installation package:" + slZipURL;
			log.error(msg);
			log.end();
			return null;
		}
		LoggingUtil.end(logger);
		
		return jbiPackage;
	}
	
	
	protected void init() 
	{
		log = new LoggingUtil(logger);
		log.start();
			
		// initialize packages loading cache
		packagesLoadedCache = new ArrayList<String>();
		
		log.end();
	}
	
	protected void shutdown() 
	{
		log.start();
			
		logger = null;
		
		packagesLoadedCache = null;
		
		autoLoader.stop();
		
		log.end();
	}
	
	// ----------------------------------------------------------------------
	//   Fractal BindingController implementation
	// ----------------------------------------------------------------------
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#listFc()
	 */
	public String[] listFc() {
		return CLIENT_INTERFACES;
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 */
	public Object lookupFc(String clientItfName)
	throws NoSuchInterfaceException {
		
		if (PLATFORM_CLASSLOADING_NAME.equals(clientItfName)) {
			return loaderSrv;
		}
		else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			return repositorySrv;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			return jmxAdmin;
		}
		else if (JBI_MESSAGING_NAME.equals(clientItfName)) {
			return messagingSrv;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String, java.lang.Object)
	 */
	public void bindFc(String clientItfName, Object serverItf)
	throws NoSuchInterfaceException, IllegalBindingException,
	IllegalLifeCycleException {
		
		if (LOGGER_NAME.equals(clientItfName)) {
			logger = (Logger) serverItf;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}
		else if (PLATFORM_CLASSLOADING_NAME.equals(clientItfName)) {
			loaderSrv = (LoaderManager) serverItf;
		}
		else if (PLATFORM_CLASSLOADING_NAME.equals(clientItfName)) {
			loaderSrv = (LoaderManager) serverItf;
		}
		else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			repositorySrv = (RepositoryService) serverItf;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			jmxAdmin = (ManagerService) serverItf;
		}
		else if (JBI_MESSAGING_NAME.equals(clientItfName)) {
			messagingSrv = (MessagingService) serverItf;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
	IllegalBindingException, IllegalLifeCycleException {
		
		if (LOGGER_NAME.equals(clientItfName)) {
			logger = null;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}
		else if (PLATFORM_CLASSLOADING_NAME.equals(clientItfName)) {
			loaderSrv = null;
		}
		else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
			repositorySrv = null;
		}
		else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			jmxAdmin = null;
		}
		else if (JBI_MESSAGING_NAME.equals(clientItfName)) {
			messagingSrv = null;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}
	
	// ----------------------------------------------------------------------
	//   Fractal LifecycleController implementation
	// ----------------------------------------------------------------------
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return (STARTED.equals(fcState) ? STARTED : STOPPED);
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {
		
		init();
		
		fcState = STARTED;
	}
	
	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {
		
		shutdown();
		
		fcState = STOPPED;
	}
	
}
