package org.objectweb.petals.jbi.management;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.management.ObjectName;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.classloader.LoaderManager;
import org.objectweb.petals.classloader.PetalsClassLoader;
import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.component.InstallationContextImpl;
import org.objectweb.petals.jbi.descriptor.ComponentDescription;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;


public class Installer implements InstallerMBean {

	/**
	 * Bootstrap classloader id prefix
	 */
	public static final String bsPrefix = "BS";

	/**
	 * Component classloader id prefix
	 */
	public static final String compPrefix = "COMP";

	
	private static final String bindingComponent = "binding-component";
	private static final String serviceEngine = "service-engine";
	
    private ManagerService jmxAdmin;
    
    private boolean installed;
    
    private LoggingUtil log;
	
	private InstallationContextImpl installContext; 
	
	private ComponentContextImpl componentContext;
	
	private ComponentDescription componentDescription;
	
	private URI installationRoot;
	
	private LoaderManager loadSrv;
	
	private Bootstrap jbiBootstrap;
	
	private ObjectName componentJmxName;
	
	
	/**
	 * Create the Installer that will manage installation / uninstallation of the Jbi component.
	 * The related bootstrap is created and initialize here.
	 * @param installContext
	 * @param componentContext
	 * @param bootstrapClassLoader
	 * @param componentClassLoader
	 * @param descriptor
	 * @param jmxAdmin
	 * @param logger
	 * @throws JBIException
	 */
	public Installer(
			URI installationRoot,
			LoaderManager loadSrv,
			ComponentContextImpl componentContext,
			ComponentDescription componentDescription, 
			ManagerService jmxAdmin,
			Logger logger) 
			throws JBIException 
	{
		super();
		
		log = new LoggingUtil(logger);
		
		log.start();
		
		this.jmxAdmin = jmxAdmin;
		
		this.componentDescription = componentDescription;
		
		this.installationRoot = installationRoot;
		
		this.componentContext = componentContext;
		
		this.loadSrv = loadSrv;
		
		// 1. create an installation context 
		// from the component context
		
		installContext = new InstallationContextImpl(componentDescription, componentContext, true);
		
		// 2. load the Bootstrap
		
		jbiBootstrap = loadBootstrap();
        
        log.end();
	}

	
	public ObjectName getInstallerConfigurationMBean() throws JBIException 
	{
		return jbiBootstrap.getExtensionMBeanName();
	}

	public String getInstallRoot() 
	{
		return componentContext.getInstallRoot();
	}

	/**
	 * Install the jbi component.
	 * Call the Bootstrap onInstall() and cleanup() methods, register the jbiComponent object,
	 * and return its jmxName.
	 * @throws JBIException the component is already installed
	 */
	public ObjectName install() throws JBIException 
	{
		log.start();
		
		if (installed) 
		{
            throw new JBIException("Component is already installed");
        }
		
		if (jbiBootstrap != null) 
        {
			jbiBootstrap.init(installContext);

        	jbiBootstrap.onInstall();
        }
        
        // Create the LifeCycle that will manage the component
        // The classpath used is the one that is set in the InstallationContext
		
		Component jbiComponent = loadComponent();
        
        ComponentLifeCycle lifecycle = new ComponentLifeCycle(componentDescription, jbiComponent, componentContext,log);
        
        
        // Register Component
                    
        if(bindingComponent.equals(componentDescription.getType()))
        {
        	componentJmxName = jmxAdmin.registerBindingComponent(lifecycle);
        } 
        else if (serviceEngine.equals(componentDescription.getType())) 
        {
        	componentJmxName = jmxAdmin.registerEngineComponent(lifecycle);
        }
    	lifecycle.setMBeanName(componentJmxName);
        
    	
        installed = true;
        
        if (jbiBootstrap != null) 
        {
        	jbiBootstrap.cleanUp();
        }
        log.end();
        
		return componentJmxName;
	}


	public boolean isInstalled() {
		
		return installed;
	}

	
	/**
	 * Uninstall the jbi component.
	 * Call the Bootstrap onUninstall() and cleanup() methods, deregister the jbiComponent object,
	 * @throws JBIException the component is already uninstalled
	 */
	public void uninstall() throws JBIException 
	{
		log.start();
		
		if (!installed) 
		{
            throw new JBIException("Component not installed");
        }
		
		if (jbiBootstrap != null)
		{
			// create an uninstallation context 
			
			InstallationContextImpl uninstallContext = new InstallationContextImpl(componentDescription, componentContext, false);
			
			jbiBootstrap.init(uninstallContext);
			
			jbiBootstrap.onUninstall();
        }
		if(bindingComponent.equals(componentDescription.getType()))
		{
        	jmxAdmin.unregisterBindingComponent(componentJmxName);
        } 
        else if (serviceEngine.equals(componentDescription.getType())) 
        {
        	jmxAdmin.unregisterEngineComponent(componentJmxName);
        }
        installed = false;
        
        if (jbiBootstrap != null)
        {
        	jbiBootstrap.cleanUp();
        }
        log.end();
	}
	
	/**
	 * Get the unique component name of this component, as assigned by the
	 * identification section of this component's installation descriptor.
	 * 
	 * @return the component name; must be non-null and non-empty.
	 */
	public String getComponentName(){
		return componentContext.getComponentName();
	}
	
	/**
	 * Create the classLoader for Bootstrap
	 * @return classLoader for Bootstrap
	 */
	private PetalsClassLoader createBootstrapClassloader()
	{
		log.call();
		
		List<String> classpath = componentDescription.getBootstrapClassPath();
		
		boolean parentFirst = JBIDescriptor.isParentFirst(
				componentDescription.getBootstrapClassLoaderDelegation());
		
		return createClassLoader(bsPrefix, classpath, parentFirst);
	}
	
	/**
	 * Create the classloader for the component
	 * Be careful, we use the classpath set in the InstallationContext.
	 * (during the installation phase, the Bootstrap can modify the 
	 * component's classpath by setting it in the InstallationContext object.)
	 * @return classloader for the component
	 */
	@SuppressWarnings("unchecked")
	private PetalsClassLoader createComponentClassloader()
	{
		log.call();
		
		List<String> classpath = (List<String>)installContext.getClassPathElements();
		
		boolean parentFirst = JBIDescriptor.isParentFirst(
				componentDescription.getComponentClassLoaderDelegation());
		
		return createClassLoader(compPrefix, classpath, parentFirst);
	}
	
	/**
	 * Create a classloader with the specified parameters.
	 * @param prefix bootstrap or component prefix
	 * @param classpath classpath of the bootstrap/component
	 * @param parentFirst delegation mode
	 * @return a classloader
	 */
	private PetalsClassLoader createClassLoader(String prefix, List<String> classpath,boolean parentFirst)
	{
		log.call();
		
		//prepare bootstrap and component classloader
		List<URL> urls = new ArrayList<URL>();
		
		try 
		{
			//TODO Find a better way of managing URI
			String installRootPath ;
			if(installationRoot.toURL().getPath().startsWith("/"))
				installRootPath = "file://"+installationRoot.toURL().getPath();
			else
				installRootPath = "file:///"+installationRoot.toURL().getPath();
			urls.add(new URL(installRootPath));
		} 
		catch (MalformedURLException e) 
		{
			// simple do nothing - JBI Package has responsibility in
			// checking the archive contents against informations
			// contained in JBI Descriptor first.
			String msg = "Unexpected error. " + e.getMessage();
			log.info(msg, e);
		}
		URL[] baseUrls = urls.toArray(new URL[] { });
		
		PetalsClassLoader bootstrapClassLoader = null;
		
		try 
		{
			bootstrapClassLoader = 
				(PetalsClassLoader) loadSrv.createClassLoader(
					prefix+componentDescription.getIdentification().getName(), 
					baseUrls, 
					classpath, 
					parentFirst);
		} 
		catch (PetalsException pe) 
		{
			// just log when some error occur
			log.error(
					"Error creating bootstrap class loader. "
					+ pe.getMessage(), pe);
			
			log.end();
			return null;
		}
		return bootstrapClassLoader;
	}
	
	/**
	 * load the component, with its own classloader
	 * @return
	 */
	private javax.jbi.component.Component loadComponent() throws JBIException
	{
		log.call();
		
		Class componentClass;
        
        javax.jbi.component.Component jbiComponent  = null;
        
        try 
        {
        	// 1. get the classloader
        	PetalsClassLoader classloader = createComponentClassloader();
        	
        	componentClass = classloader.loadClass(componentDescription.getComponentClassName());
        
        	if (componentClass == null)
        	{
                throw new JBIException("Could not find component class: " + componentDescription.getComponentClassName());
            }
        	jbiComponent = (javax.jbi.component.Component) componentClass.newInstance();
        }
        catch (ClassNotFoundException e) 
        {
        	throw new JBIException("Class not found: " + componentDescription.getComponentClassName(),e);
        }
        catch (InstantiationException e) 
        {
        	throw new JBIException("Could not instantiate : " + componentDescription.getComponentClassName(),e);
        }
        catch (IllegalAccessException e) 
        {
        	throw new JBIException("Illegal access on: " + componentDescription.getComponentClassName(),e);
        }
        return jbiComponent;
	}

	/**
	 * load the bootstrap, with its own classloader
	 * @return
	 */
	private Bootstrap loadBootstrap() throws JBIException
	{
		log.call();
		
		Class bootstrapClass;
        
		Bootstrap jbiBootstrap  = null;
        
        try 
        {
        	// 1. get the classloader
        	PetalsClassLoader classloader = createBootstrapClassloader();
        	
        	bootstrapClass = classloader.loadClass(componentDescription.getBootstrapClassName());
        
        	if (bootstrapClass == null)
        	{
                throw new JBIException("Could not find bootstrap class: " + componentDescription.getBootstrapClassName());
            }
        	jbiBootstrap = (Bootstrap) bootstrapClass.newInstance();
        }
        catch (ClassNotFoundException e) 
        {
        	throw new JBIException("Class not found: " + componentDescription.getBootstrapClassName(),e);
        }
        catch (InstantiationException e) 
        {
        	throw new JBIException("Could not instantiate : " + componentDescription.getBootstrapClassName(),e);
        }
        catch (IllegalAccessException e) 
        {
        	throw new JBIException("Illegal access on: " + componentDescription.getBootstrapClassName(),e);
        }
        return jbiBootstrap;
	}
	
}
