/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005  EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: LifeCycle.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.management;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.component.event.StateChangeFailedEvent;
import org.objectweb.petals.jbi.component.event.StateChangedEvent;
import org.objectweb.petals.jbi.descriptor.ComponentDescription;
import org.objectweb.petals.util.LoggingUtil;


/**
 * Implementation of the <code>LifeCycle</code> mbean.
 * This lifecycle mbean sends notifications about changes of state.
 * @author Adrien LOUIS - EBM WebSourcing
 */
public class LifeCycle implements LifeCycleMBean, NotificationBroadcaster
{
	
	private ComponentDescription componentDescription;
	
    protected Component component;
    
    private long sequenceNumber = 0;
    
    protected LoggingUtil log;
    
    protected Object activitySynchronizer;

    protected boolean initialized;
    
    protected String state=null;
    
    private ObjectName mBeanName;
    
    private ComponentContextImpl context;
    
    public static String INITIALIZING = "Initializing";
    public static String STARTING = "Starting";
    public static String STOPPING = "Stopping";
    public static String SHUTDOWNING = "ShutDowning";
    
    
    public LifeCycle(ComponentDescription componentDescription, Component component, ComponentContextImpl context, LoggingUtil log)
    {
    	this.log = log;
    	
    	log.call();
    
    	this.componentDescription=componentDescription;
    	
    	this.component = component;
        
    	this.context = context;
        
    	activitySynchronizer = new Object();
    	
    	state = SHUTDOWN;
    	
    	initialized = false;
    	
    	// set the jbiComponent ref to the contextImpl
    	if(context!=null){
    		context.setComponent(component);
    	}
    }
    
    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState()
    {
        return state;
    }
    
    /**
     * Set the <code>State</code> reference of the current state.
     * An <code>LifeCycleStateEvent</code> is thrown.
     * @param state
     */
    public synchronized void setState(String state)
    {
        stateChanged(new StateChangedEvent(this,this.state,state));
        
        this.state = state;
    }
    
    public ObjectName getMBeanName() {
		return mBeanName;
	}

	public void setMBeanName(ObjectName beanName) {
		mBeanName = beanName;
	}
	
	public String getName()
	{
		return componentDescription.getIdentification().getName();
	}

	
	//////////////////////////////////////////////
	// LifeCycle implementation
	//////////////////////////////////////////////
	
	/**
     * Shutdown the component.
     * A notification will be thrown in case of runtimeException or incoherent state.
     * @see javax.jbi.management.LifeCycleMBean#shutdown()
     * @throws JBIException in case of runtimeException or incoherent state.
     */
    public void shutDown() throws JBIException
    {
    	log.start();

    	if(STOPPED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				setState(SHUTDOWNING);

    				component.getLifeCycle().shutDown();
            
    				setState(SHUTDOWN);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the shutdown.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Component can not be shutdowned in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
    }
    
    
    /**
     * Start the component.
     * A notification will be thrown in case of runtimeException or incoherent state.
     * @see javax.jbi.management.LifeCycleMBean#start()
     * @throws JBIException in case of runtimeException or incoherent state.
     */
    public void start() throws JBIException
    {
    	log.start();

    	if(SHUTDOWN.equals(state) || STOPPED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				if(! initialized)
    				{
    					setState(INITIALIZING);

    					component.getLifeCycle().init(context);
                
    					initialized = true;
    				}
    				setState(STARTING);

    				component.getLifeCycle().start();
            
    				setState(STARTED);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the start.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Component can not be started in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
    }
    
    /**
     * Stop the component.
     * A notification will be thrown in case of runtimeException or incoherent state.
     * @see javax.jbi.management.LifeCycleMBean#stop()
     * @throws JBIException in case of runtimeException or incoherent state.
     */
    public void stop() throws JBIException
    {
    	log.start();

    	if(STARTED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				setState(STOPPING);

    				component.getLifeCycle().stop();
            
    				setState(STOPPED);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the stop.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Component can not be stoped in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
    }
    
    
    
    
    
    //
    // Notifications support
    //
    
    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport() ;
    
    public void addNotificationListener(NotificationListener arg0,
            NotificationFilter arg1, Object arg2)
    throws IllegalArgumentException
    {
        notifSupport.addNotificationListener(arg0, arg1, arg2);
    }
    
    
    public MBeanNotificationInfo[] getNotificationInfo()
    {
        String[] types = new String[] { 
                AttributeChangeNotification.ATTRIBUTE_CHANGE 
        }; 
        
        String name = AttributeChangeNotification.class.getName(); 
        
        String description = "Notification about a change of the lifeCycle state or a state change failure."; 
        
        MBeanNotificationInfo info = 
            new MBeanNotificationInfo(types, name, description); 
        
        return new MBeanNotificationInfo[] {info}; 
    } 
    
  
    
    public void removeNotificationListener(NotificationListener arg0)
    throws ListenerNotFoundException
    {
        notifSupport.removeNotificationListener(arg0);
    }
    
    //////////////////////////////////////////////
    // ElementListener implementation
    //////////////////////////////////////////////
    
    /**
     * <i>stateChanged</i> is called when the state of an element (JBI Component or SystemService)
     * changes.
     * A JMX-notification is thrown.
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChanged(StateChangedEvent event)
    {
    	log.call();
        Notification n = 
            new AttributeChangeNotification(
                    mBeanName,
                    sequenceNumber++,
                    System.currentTimeMillis(),
                    "State changed : "+event.getOldState()+" -> "+event.getNewState(),
                    "state",
                    "String",
                    event.getOldState(),
                    event.getNewState()); 
        
        notifSupport.sendNotification(n); 
    }
    
    /**
     * <i>stateChangeFailed</i>> is called when the change of state of an element (JBI Component or SystemService)
     * fails.
     * A JMX-notification throws the relative JBIException, .
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChangeFailed(StateChangeFailedEvent event)
    {
    	log.call();
        Notification n = 
            new AttributeChangeNotification(
                    mBeanName,
                    sequenceNumber++,
                    System.currentTimeMillis(),
                    "State change failed : "+event.getException().getMessage(),
                    "exception",
                    "String",
                    null,
                    event.getException().toString()); 
        
        notifSupport.sendNotification(n); 
    }

	public Component getComponent() {
		return component;
	}
    
    
    
}
