/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005  EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: Manager.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.management;

import static org.objectweb.petals.util.InterfaceNamesHelper.JMXAGENT_MBEANSERVERADMIN_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

import java.util.Map.Entry;




/**
 * <code>Manager</code> is the main class for JMX management.
 * It is the entry point for register and unregister elements in the JMX environment
 * (elements are JBI binding/engine components and systemServices
 * components).
 * Related MBeans are created when registering and unregistering elements.
 * 
 * <code>Manager</code> owns the JMX management services offered by the JBI container 
 * to administrative clients. 
 * Management services are :
 * - AdminService (referenced in the JMX server as "admin");
 * - InstallationService (referenced in the JMX server as "installation");
 * - DeploymentService (referenced in the JMX server as "deployment");
 * 
 * <code>Manager</code> is the central point for accessing the mBeans associated 
 * to the JBI container elements.
 * 
 * <code>ManagementException</code> exceptions are thrown when
 * problems for accessing the JMX server occur.
 *  @author Adrien LOUIS - EBM WebSourcing
 */
// TODO refactor to get System Services from Fractal, and BC and SE from JMX server if possible
public class ManagerImpl implements ManagerService, BindingController, LifeCycleController 
{
    
    private static final String[] CLIENT_INTERFACES = new String[] {
        JMXAGENT_MBEANSERVERADMIN_NAME };
    
	/**
	 * Component lifecycle state.
	 */
	private String fcState;
	
    protected MBeanNamesImpl mBeanNames;
    
    protected MBeanServer mBeanServer;
    
    /**
     * Monolog logger
     */
    protected Logger logger;
    
    /**
	 * Logger wrapper
	 */
	private LoggingUtil log;
    
    
    protected Map<ObjectName,ComponentLifeCycle> sync_bindings;
    
    protected Map<ObjectName,ComponentLifeCycle> sync_engines;
    
    protected Map<ObjectName,LifeCycle> sync_systemServices;
    
    protected Map<String,ServiceAssemblyLifeCycle> sync_serviceAssemblyLifeCycles;
    
    protected Map<ObjectName,Installer> sync_installers;
    
    private static final String DEPLOYMENT_NAME = "deployment";
    
    private static final String INSTALLATION_NAME = "installation";
    
    protected ObjectName directoryName;
    
    protected ObjectName transporterName;
    
    protected ObjectName itineraryName;
    
    protected ObjectName addressName;
    
    protected ObjectName adminName;
    
    protected ObjectName deploymentName;
    
    protected ObjectName installationName;


	/**
	 * Default constructor.
     * Creates a <code>Manager</code> instance.
     * JMX server is created. 
     *
     */
    public ManagerImpl()
    {    
        
        mBeanNames = new MBeanNamesImpl("org.objectweb.petals");
        
        sync_bindings = Collections.synchronizedMap(new HashMap<ObjectName,ComponentLifeCycle>());
        
        sync_engines = Collections.synchronizedMap(new HashMap<ObjectName,ComponentLifeCycle>());
        
        sync_systemServices = Collections.synchronizedMap(new HashMap<ObjectName,LifeCycle>());
        
        sync_serviceAssemblyLifeCycles = Collections.synchronizedMap(new HashMap<String,ServiceAssemblyLifeCycle>());
        
        sync_installers = Collections.synchronizedMap(new HashMap<ObjectName,Installer>());
        
        deploymentName = mBeanNames.createSystemComponentMBeanName(DEPLOYMENT_NAME);
        
        installationName = mBeanNames.createSystemComponentMBeanName(INSTALLATION_NAME);
        
    }
    
    /**
     * Return the <code>MBeanNamesImpl</code> element which 
     * allows the creation of mBeans names.
     * @return
     */
    public MBeanNamesImpl getMBeanNames()
    {
        return mBeanNames;
    }
    
    /**
     * Return the JMX server used by the JBI container.
     * @return
     */
    public MBeanServer getMBeanServer()
    {
        return mBeanServer;
    }
    
    /**
     * Return the synchronized map of binding components.
     * @return
     */
    public Map<ObjectName,ComponentLifeCycle> getSync_bindings()
    {
        return sync_bindings;
    }

    /**
     * Return the synchronized map of engine components.
     * @return
     */
    public Map<ObjectName,ComponentLifeCycle> getSync_engines()
    {
        return sync_engines;
    }

    /**
     * Return the synchronized map of system services components. 
     * @return
     */
    public Map<ObjectName,LifeCycle> getSync_systemServices()
    {
        return sync_systemServices;
    }
    
    /**
     * Return the synchronized map of deployed service assembly.
     * @return
     */
    public Map<String, ServiceAssemblyLifeCycle> getSync_serviceAssemblyLifeCycles() {
		return sync_serviceAssemblyLifeCycles;
	}
    
    /**
     * Return the synchronized map of installers.
     * @return
     */
    public Map<ObjectName, Installer> getSync_installers() {
		return sync_installers;
	}

	/**
     * Register a ServiceAssemblyLifeCycle into the a synchronized Map<String,ServiceAssemblyLifeCycle>
     * @param saName
     * @param serviceAssemblyLifeCycle
     */
    public void registerServiceAssembly(String saName, ServiceAssemblyLifeCycle serviceAssemblyLifeCycle) {
    	Map<String,ServiceAssemblyLifeCycle> sas = getSync_serviceAssemblyLifeCycles();
    	synchronized (sas) {
    		sas.put(saName, serviceAssemblyLifeCycle);
		}
    }
    
    /**
     * Unregister the ServiceAssembly life cycle from the sync map
     * @throws ManagementException
     */
    public void unregisterServiceAssembly(String saName) throws ManagementException
    {
    	Map sas = getSync_serviceAssemblyLifeCycles();
        
        synchronized(sas)
        {
            try
            {
                sas.remove(saName);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
    }
    
    /**
     * Return the service assembly associated with this name.
     * @param saName
     * @return The related ServiceAssemblyLifeCycle, or null
     */
    public ServiceAssemblyLifeCycle getServiceAssemblyByName(String saName)
    {
    	log.call();
    	
    	ServiceAssemblyLifeCycle result = null;
    	
    	Map<String,ServiceAssemblyLifeCycle> sas = this.getSync_serviceAssemblyLifeCycles();
        
    	synchronized(sas)
        {
            result = sas.get(saName);
        }
    	return result;
    }
    
    
    //////////////////////////////
    // register components
    //////////////////////////////

	/**
     * Register a mbean in the JMX server representing 
     * the specified binding component.
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerBindingComponent(ComponentLifeCycle lifeCycle) throws ManagementException
    {
        
        ObjectName name = mBeanNames.createBindingComponentMBeanName(lifeCycle.getName());
        
        Map<ObjectName,ComponentLifeCycle> bindings = getSync_bindings();
        
        synchronized(bindings)
        {
            try
            {
                mBeanServer.registerMBean(lifeCycle, name);
                
                bindings.put(name, lifeCycle);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
        return name;
    }
    
    /**
     * Unregister the binding component of the JMX server
     * @throws ManagementException
     */
    public void unregisterBindingComponent(ObjectName name) throws ManagementException
    {
    	Map bindings = getSync_bindings();
        
        synchronized(bindings)
        {
            try
            {
                mBeanServer.unregisterMBean(name);
                
                bindings.remove(name);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
    }
    
	/**
     * Register an installer mbean in the JMX server representing 
     * the specified component installer.
     * @param installerMBean
     * @throws ManagementException
     */
    public ObjectName registerInstaller(Installer installerMBean) throws ManagementException
    {
        
    	ObjectName installerMBeanName = mBeanNames.createInstallerMBeanName(installerMBean.getComponentName());

        Map<ObjectName,Installer> installers = getSync_installers();
        
        synchronized(installers)
        {
            try
            {
        		mBeanServer.registerMBean(installerMBean, installerMBeanName);
                
                installers.put(installerMBeanName, installerMBean);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
        return installerMBeanName;
    }
    
    /**
     * Unregister the component installer of the JMX server
     * @throws ManagementException
     */
    public void unregisterInstaller(ObjectName name) throws ManagementException
    {
    	Map<ObjectName,Installer> installers = getSync_installers();
        
        synchronized(installers)
        {
            try
            {
                mBeanServer.unregisterMBean(name);
                
                installers.remove(name);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
    }
    
    /**
     * Register a mbean in the JMX server representing 
     * the specified engine component
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerEngineComponent(ComponentLifeCycle lifeCycle) throws ManagementException
    {
        ObjectName name = mBeanNames.createEngineComponentMBeanName(lifeCycle.getName());
        
        Map<ObjectName,ComponentLifeCycle> engines = getSync_engines();
        
        synchronized(engines)
        {
            try
            {
                mBeanServer.registerMBean(lifeCycle, name);
                
                engines.put(name, lifeCycle);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
        return name;
    }
    
    /**
     * Unregister the engine component of the JMX server
     * @throws ManagementException
     */
    public void unregisterEngineComponent(ObjectName name) throws ManagementException
    {
    	Map engines = getSync_engines();
        
        synchronized(engines)
        {
            try
            {
                mBeanServer.unregisterMBean(name);
                
                engines.remove(name);
            }
            catch (Exception e)
            {
                throw new ManagementException(e);
            }
        }
    }
    
    /**
     * @see org.objectweb.petals.jbi.management.ManagerService#getComponentByName(java.lang.String)
     */
    public LifeCycle getComponentByName(String name)
    {
        log.call();
        
        LifeCycle on = null;
        
        // search in bindings
        
        on = getBindingComponentByName(name);
        
        if(on == null)
        {
        	// search in engines
        
        	on = getEngineComponentByName(name);
        }        	
        return on;
    }
   
    /**
     * Return the engine component associated with this name.
     * @param name
     * @return The related objectName, or null
     */
    private LifeCycle getEngineComponentByName(String name)
    {
    	log.call();
    	
    	LifeCycle result = null;
    	
    	Map engines = this.getSync_engines();
        
    	synchronized(engines)
        {
            for (Iterator iter = engines.values().iterator(); iter.hasNext() && result == null;)
            {
                LifeCycle lifeCycle = (LifeCycle) iter.next();
                
                if (lifeCycle.getName().equals(name))
                {
                	result = lifeCycle;
                }
            }
        }
    	return result;
    }
    
    /**
     * Return the binding component associated with this name.
     * @param name
     * @return The related objectName, or null
     */
    private LifeCycle getBindingComponentByName(String name)
    {
    	log.call();
    	
    	LifeCycle result = null;
    	
    	Map bindings = this.getSync_bindings();
        
    	synchronized(bindings)
        {
            for (Iterator iter = bindings.values().iterator(); iter.hasNext() && result == null;)
            {
                LifeCycle lifeCycle = (LifeCycle) iter.next();
                
                if (lifeCycle.getName().equals(name))
                {
                	result = lifeCycle;
                }
            }
        }
    	return result;
    }
    
    /**
     * Return the component installer associated with this componant name.
     * @param name the component name
     * @return The related installer, or null
     */
    @SuppressWarnings("unchecked")
	public ObjectName getInstallerByName(String name)
    {
    	log.call();
    	
    	ObjectName result = null;
    	
    	Map<ObjectName,Installer> installers = getSync_installers();
        
    	synchronized(installers)
        {
            for (Iterator iter = installers.entrySet().iterator(); iter.hasNext() && result == null;)
            {
            	Entry<ObjectName,Installer> entry = (Entry<ObjectName,Installer>) iter.next();
            	ObjectName objName = entry.getKey();
            	Installer installer = entry.getValue();
                
                if (installer.getComponentName().equals(name))
                {
                	result = objName;
                }
            }
        }
    	return result;
    }
    
	// ----------------------------------------------------------------------
	//   Fractal BindingController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#listFc()
	 */
	public String[] listFc() {
		return CLIENT_INTERFACES;
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 */
	public Object lookupFc(String clientItfName)
			throws NoSuchInterfaceException {

		if (JMXAGENT_MBEANSERVERADMIN_NAME.equals(clientItfName)) {
			return mBeanServer;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String, java.lang.Object)
	 */
	public void bindFc(String clientItfName, Object serverItf)
			throws NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException {
		if (LOGGER_NAME.equals(clientItfName)) {
			logger = (Logger) serverItf;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}
		else if (JMXAGENT_MBEANSERVERADMIN_NAME.equals(clientItfName)) {
			mBeanServer = (MBeanServer) serverItf;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {

		if (LOGGER_NAME.equals(clientItfName)) {
			logger = null;
		}
		else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}else if (JMXAGENT_MBEANSERVERADMIN_NAME.equals(clientItfName)) {
			mBeanServer = null;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	// ----------------------------------------------------------------------
	//   Fractal LifecycleController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return (STARTED.equals(fcState) ? STARTED : STOPPED);
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {

		log = new LoggingUtil(logger);
		
		fcState = STARTED;
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {

		fcState = STOPPED;
	}

}

