/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationService.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management;

import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.component.ServiceUnitManager;

import org.objectweb.petals.jbi.component.event.StateChangeFailedEvent;
import org.objectweb.petals.jbi.component.event.StateChangedEvent;
import org.objectweb.petals.jbi.descriptor.ServiceAssembly;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This class manage the life cycle of deployed service assemblies
 * @author Olivier Fabre - EBM Websourcing
 *
 */
public class ServiceAssemblyLifeCycle extends LifeCycle {
	
    private Map<String,ServiceUnitLifeCycle> sync_sus;
    
    private URI installationRoot;
    
    private ServiceAssembly serviceAssembly;
    
    private RepositoryService repositoryService;
	
	public ServiceAssemblyLifeCycle(LoggingUtil log, ServiceAssembly serviceAssembly, 
									RepositoryService repositoryService, URI installationRoot) {
		super(null, null, null, log);
		this.installationRoot = installationRoot;
		this.serviceAssembly = serviceAssembly;
		this.repositoryService = repositoryService;
		this.sync_sus = Collections.synchronizedMap(new HashMap<String,ServiceUnitLifeCycle>());
	}

	@Override
	public void shutDown() throws JBIException {
    	log.start();

    	if(STOPPED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				setState(SHUTDOWNING);

    				Map<String,ServiceUnitLifeCycle> sus = this.getSync_sus();
    		        
    		    	synchronized(sus)
    		        {
    		            for (Iterator iter = sus.values().iterator(); iter.hasNext();)
    		            {
    		            	ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter.next();
    		                lifeCycle.shutDown();
    		            }
    		        }
            
    				setState(SHUTDOWN);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the shutdown.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Service Assembly can not be shutdowned in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
	}

	@Override
	public void start() throws JBIException {
		log.start();

    	if(SHUTDOWN.equals(state) || STOPPED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				setState(STARTING);
    				
    				Map<String,ServiceUnitLifeCycle> sus = this.getSync_sus();
    		        
    		    	synchronized(sus)
    		        {
    		            for (Iterator iter = sus.values().iterator(); iter.hasNext();)
    		            {
    		            	ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter.next();
    		                lifeCycle.start();
    		            }
    		        }
            
    				setState(STARTED);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the start.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Service Assembly can not be started in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
	}


	@Override
	public void stop() throws JBIException {
		log.start();

    	if(STARTED.equals(state))
    	{
    		synchronized(activitySynchronizer)
    		{
    			try
    			{
    				setState(STOPPING);

    				Map<String,ServiceUnitLifeCycle> sus = this.getSync_sus();
    		        
    		    	synchronized(sus)
    		        {
    		            for (Iterator iter = sus.values().iterator(); iter.hasNext();)
    		            {
    		            	ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter.next();
    		                lifeCycle.stop();
    		            }
    		        }
            
    				setState(STOPPED);
    			}
    			catch (RuntimeException re)
    			{
    				setState(UNKNOWN);
    				
    				JBIException e = new JBIException("A runtime exception occured during the stop.",re);
    				
    				log.error(e);
    				
    				stateChangeFailed(new StateChangeFailedEvent(this,e));
    				
    				throw e;
    			}    
    		}
    	}
    	else
    	{
    		JBIException e = new JBIException("The Service Assembly can not be stoped in this state.");
			
    		log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
    		
    		throw e;
    	}
        log.end();
	}
	
	public void undeploy() throws JBIException {
		log.start();
		// Service assembly must be in the shutdown state
		if (SHUTDOWN.equals(state)){
			// Undeploy all service units
			Map<String,ServiceUnitLifeCycle> sus = getSync_sus();
			synchronized(sus)
			{
				for (Iterator iter = sus.values().iterator(); iter.hasNext();)
				{
					ServiceUnitLifeCycle lifeCycle = (ServiceUnitLifeCycle) iter.next();
					
					ServiceUnitManager sum = lifeCycle.getServiceUnitManager();
					if (sum!=null){
						sum.undeploy(lifeCycle.getSuName(), lifeCycle.getServiceUnitRootPath());
					}
				}
			}
			// Remove the Service assembly folder.
			repositoryService.removePackage(serviceAssembly.getIdentification().getName());
		}else {
			JBIException e = new JBIException("The Service Assembly can not be undeployed in this state.");
			
			log.error(e);
			
			stateChangeFailed(new StateChangeFailedEvent(this,e));
			
			throw e;
		}
		log.end();
	}
	
    /**
     * Register a ServiceAssemblyLifeCycle into the a synchronized Map<String,ServiceAssemblyLifeCycle>
     * @param saName
     * @param serviceAssemblyLifeCycle
     */
    public void registerSUIntoSA(String suName, ServiceUnitLifeCycle serviceUnitLifeCycle) {
    	Map<String,ServiceUnitLifeCycle> sus = getSync_sus();
    	synchronized (sus) {
    		sus.put(suName, serviceUnitLifeCycle);
		}
    }

    /**
     * Return the synchronized map of deployed service units.
     * @return
     */
	public Map<String, ServiceUnitLifeCycle> getSync_sus() {
		return sync_sus;
	}

	public ServiceAssembly getServiceAssembly() {
		return serviceAssembly;
	}

	public URI getInstallationRoot() {
		return installationRoot;
	}

	/**
	 * Do nothing because there's no Service Assembly MBean registered in JMX !
	 */
	@Override
	public synchronized void stateChanged(StateChangedEvent event) {
		
	}

	/**
	 * Do nothing because there's no Service Assembly MBean registered in JMX !
	 */
	@Override
	public synchronized void stateChangeFailed(StateChangeFailedEvent event) {
		
	}

}
