/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 � EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AutoLoaderImpl.java,v 1.2 2006/04/07 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.management.autoload;

import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_INSTALLATION_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_DEPLOYMENT_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_JMXADMIN_NAME;


import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Timer;

import javax.jbi.management.DeploymentServiceMBean;
import javax.jbi.management.InstallationServiceMBean;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.jbi.management.ManagerService;
import org.objectweb.petals.util.FileUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is used for automaticly loading components 
 * which are placed in the "autoLoad" directory.
 * It periodicly scan the directory and check if
 * new components are presents.
 * The InstallationService is called for each of those components.
 * The whole process of installation is done (install + start)
 * After being threat, the archives of the components are removed. 
 * A file "fileName.success" will be created if success
 * A file "fileName.failed" will be created otherwise, containing the related exception
 * 
 * @author alouis
 *
 */
public class AutoLoaderImpl implements AutoLoader,BindingController, LifeCycleController{
	
	private static final String[] CLIENT_INTERFACES = 
		new String[] {JBI_DEPLOYMENT_NAME,JBI_INSTALLATION_NAME,JBI_JMXADMIN_NAME};
	
	/**
	 * Component LifeCycle state.
	 */
	private String fcState;

	
	private static String autoinstall = "autoinstall";
	private static String autodeploy = "autodeploy";
	private static String autoshare = "autoshare";
	private static String work  = "work";
	private long scanPeriod = 1000L;
	
	private static File autoinstallDirectory = null;
	private static File autodeployDirectory = null;
	private static File autoshareDirectory = null;
	private static File workDirectory = null;
	
	private Timer directoryScanTimer = null;
	
	private InstallationServiceMBean installationService = null;
	private DeploymentServiceMBean deploymentService = null;
	
	private ManagerService managerService = null;
	
	private LoggingUtil log;
	
	/**
	 * init the autoloader : init the directories 
	 */
	public void init() throws IOException
	{
		log.call();
		autoinstallDirectory = new File(SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator + autoinstall);
		autodeployDirectory = new File(SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator + autodeploy);
		autoshareDirectory = new File(SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator + autoshare);
		workDirectory = new File(SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator + work);
		
		checkValidiyOfdirectory(autoinstallDirectory,false);
		checkValidiyOfdirectory(autodeployDirectory,false);
		checkValidiyOfdirectory(autoshareDirectory,false);
		checkValidiyOfdirectory(workDirectory,true);
		
		//for the moment, clean the repository
		File repository = new File(SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator + "repository");
		checkValidiyOfdirectory(repository ,true);
		
		
	}
	
	/**
	 * start the autoloader : launch the scan 
	 */
	public void start()
	{
		log.call();
		
		directoryScanTimer = new Timer();
		DirectoryScanner scan = new DirectoryScanner(this,autoinstallDirectory,autodeployDirectory,autoshareDirectory);
		
		directoryScanTimer.schedule(scan,scanPeriod,scanPeriod);
	}
	
	/**
	 * Stop the autoload of the components.
	 * TODO : synchronization, does the autoload must wait before the end of an installation before closing?
	 */
	public void stop()
	{
		log.call();
		
		if(directoryScanTimer != null)
		{
			directoryScanTimer.cancel();
		}
	}
	
	/**
	 * friendly method for the directoryscanner
	 * @param newDeployers
	 */
	void deploy(File[] newDeployers) {
		log.call();
		
		if(newDeployers != null)
		{
			for (int i = 0; i < newDeployers.length; i++) 
			{
				File file = newDeployers[i];
				performDeploy(file);
			}
		}
		
	}
	
	/**
	 * friendly method for the directoryscanner
	 * @param newInstallers
	 */
	void install(File[] newInstallers) 
	{
		log.call();
		
		if(newInstallers != null)
		{
			for (int i = 0; i < newInstallers.length; i++) 
			{
				File file = newInstallers[i];
				performInstall(file);
			}
		}
	}
	
	/**
	 * friendly method for the directoryscanner
	 * @param newInstallers
	 */
	void share(File[] newInstallers) 
	{
		log.call();
		
		if(newInstallers != null)
		{
			for (int i = 0; i < newInstallers.length; i++) 
			{
				File file = newInstallers[i];
				performShare(file);
			}
		}
	}
	
	/**
	 * Install the specified Archive in the JBI container
	 * and start it .
	 * Any exception will cause the abortion of the process.
	 * A file "fileName.success" will be created if success
	 * A file "fileName.failed" will be created otherwise, containing the related exception
	 * @param f
	 */
	private void performInstall(File f)
	{
		log.call();
		
		if(f != null)
		{
			File workFile = null;
			
			try {
				
				// move file to Work directory
				
				workFile = new File(workDirectory,f.getName());
				
				f.renameTo(workFile);

				f=null;
				
				if(workFile.exists())
				{
					MBeanServer jmxServer = managerService.getMBeanServer();
					
					//TODO Find a better way of managing URI
					String absPath = workFile.getAbsolutePath();
					String archiveURI ;
					if(absPath.startsWith("/"))
						archiveURI= "file://"+absPath;
					else
						archiveURI= "file:///"+absPath;
					
					ObjectName newInstallerName = installationService.loadNewInstaller(archiveURI);
					
					Object[] nullObjects= new Object[0];
					String[] nullStrings= new String[0];
					
					Object result = jmxServer.invoke(newInstallerName,"install",nullObjects,nullStrings);
					
					if(result != null && result instanceof ObjectName)
					{
						ObjectName componentName = (ObjectName)result;
						
						jmxServer.invoke(componentName,"start",nullObjects,nullStrings);
						
						new File(workFile.getCanonicalPath()+".success").createNewFile();
						
						workFile.delete();
					}
					else
					{
						throw new Exception("error during installation");
					}
				}
			} 
			catch (Throwable e) 
			{
				try {
					String workFilePath = workFile.getCanonicalPath();
					
					File failed = new File(workFilePath+".failed");
					failed.createNewFile();
					PrintWriter w = null;
					w = new PrintWriter(failed);
					e.printStackTrace(w);
					w.flush();
					w.close();
					//TODO log
					e.printStackTrace();
				} catch (IOException e1) {
					//TODO log
					e.printStackTrace();
				}
			}
		}
	}
	
	
	/**
	 * Install the specified shared library in the JBI container
	 * Any exception will cause the abortion of the process.
	 * A file "fileName.success" will be created if success
	 * A file "fileName.failed" will be created otherwise, containing the related exception
	 * @param f
	 */
	private void performShare(File f)
	{
		log.call();
		
		if(f != null)
		{
			File workFile = null;
			
			try {
				
				// move file to Work directory
				
				workFile = new File(workDirectory,f.getName());
				
				f.renameTo(workFile);

				f=null;
				
				if(workFile.exists())
				{
					String archiveURI=workFile.toURI().toString();
					
					installationService.installSharedLibrary(archiveURI);
				}
			} 
			catch (Throwable e) 
			{
				try {
					String workFilePath = workFile.getCanonicalPath();
					
					File failed = new File(workFilePath+".failed");
					failed.createNewFile();
					PrintWriter w = null;
					w = new PrintWriter(failed);
					e.printStackTrace(w);
					w.flush();
					w.close();
					//TODO log
					e.printStackTrace();
				} catch (IOException e1) {
					//TODO log
					e.printStackTrace();
				}
			}
		}
	}
	
	
	/**
	 * Deploy the specified Archive in the JBI container
	 * and start it .
	 * Any exception will cause the abortion of the process.
	 * A file "fileName.success" will be created if success
	 * A file "fileName.failed" will be created otherwise, containing the related exception
	 * @param f
	 */
	private void performDeploy(File f)
	{
		log.call();
		
		if(f != null)
		{
			File workFile = null;
			
			try {
				
				// move file to Work directory
				
				workFile = new File(workDirectory,f.getName());
				
				f.renameTo(workFile);

				f=null;
				
				if(workFile.exists())
				{
					String archiveURI=workFile.toURI().toString();
					
					String deploymentState = deploymentService.deploy(archiveURI);
					
					//TODO : the deploy() method return an XML document. for the moment it is the assemblyName 
					deploymentService.start(deploymentState);
				}
			} 
			catch (Throwable e) 
			{
				try {
					String workFilePath = workFile.getCanonicalPath();
					
					File failed = new File(workFilePath+".failed");
					failed.createNewFile();
					PrintWriter w = null;
					w = new PrintWriter(failed);
					e.printStackTrace(w);
					w.flush();
					w.close();
					//TODO log
					e.printStackTrace();
				} catch (IOException e1) {
					//TODO log
					e.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * Check if the specified File is a directory and is created.
	 * If it does not exist, create it.
	 * @param dir
	 * @param clean delete directory content or not
	 */
	private void checkValidiyOfdirectory(File dir,boolean clean) throws IOException
	{
		if(dir != null)
		{
			// parameter is not null
			if(! dir.exists())
			{
				// directory does not exist, create it
				if(!dir.mkdirs())
				{
					// creation failed
					throw new IOException("can not create the following directory : "+dir.getCanonicalPath());
				}
			}
			else
			{
				// file already exist, check it is a directory
				if(! dir.isDirectory())
				{
					// file is not a directory
					throw new IOException(dir.getCanonicalPath()+" already exists and is not a directory.");
				}
				else
				{
					// clean the directory
					if(clean)
					{
						FileUtil.removeDirectory(dir);
						dir.mkdirs();
					}
				}
			}
		}
		else
		{
			throw new FileNotFoundException("the specified directory is null");
		}
	}
	
	
	//  ----------------------------------------------------------------------
	//   Fractal BindingController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#listFc()
	 */
	public String[] listFc() {
		return CLIENT_INTERFACES;
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 */
	public Object lookupFc(String clientItfName)
			throws NoSuchInterfaceException {
		if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
			return installationService;
		}else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
			return deploymentService;
		}else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			return managerService;
		}else {
			throw new NoSuchInterfaceException(clientItfName);
		}		
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String, java.lang.Object)
	 */
	public void bindFc(String clientItfName, Object serverItf)
			throws NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException {

		if (LOGGER_NAME.equals(clientItfName)) {
			log = new LoggingUtil((Logger) serverItf);
		}else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}else if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
			installationService = (InstallationServiceMBean)serverItf;
		}else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
			deploymentService = (DeploymentServiceMBean)serverItf;
		}else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			managerService = (ManagerService)serverItf;
		}else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {
		if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
			installationService = null;
		}else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
			deploymentService = null;
		}else if (JBI_JMXADMIN_NAME.equals(clientItfName)) {
			managerService = null;
		}else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	// ----------------------------------------------------------------------
	//   Fractal LifecycleController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return (STARTED.equals(fcState) ? STARTED : STOPPED);
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {

		try {
			init();
		} catch (IOException e) {
			log.error(e);
			throw new IllegalLifeCycleException(e.getMessage());
		}
		start();
    	fcState = STARTED;
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {
		stop();
		fcState = STOPPED;
	}
	
}
