/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIPackage.java 76 2006-01-17 17:15:44Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.packaging;

import java.net.URI;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;

/**
 * The JBIPackage interface encapsulates installation and deployment packages
 * usual functions, providing additional validation and helper methods. An
 * enhanced validation is added by checking JBI Descriptors and package archive
 * for referenced entries, complementing a simple descriptor schema validation.
 * <p>
 * This interface must return the loaded, parsed and validated JBI Descriptor
 * corresponding to the JBI Package ZIP archive file. 
 * <p>
 * The installation package is a ZIP archive file, which has contents that are
 * opaque to JBI except for one installation descriptor that MUST be named and
 * located as follows:
 * <pre>
 *     /META-INF/jbi.xml
 * </pre>
 * <p>
 * The deployment package consists of a deployment descriptor, and one or more
 * service unit archives, all contained within a ZIP archive file. The
 * file/directory structure of this archive is as follows:
 * <ul>
 *   <li><b>/META-INF/jbi.xml</b>. This contains the deployment descriptor,
 *   which MUST conform to the deployment descriptor schema.</li>
 *   <li><b>/<i>{artifacts-file-name.zip}</i></b>. This is the name of one of
 *   the service unit archives, as given in one of the deployment descriptor's
 *   target elements.</li>
 *   <li><b>/<i>{artifacts-file-name-N.zip}</i></b>. As many service unit
 *   archives as are needed provide all such archives mentioned in the
 *   deployment descriptor <i>jbi.xml</i>.</li>
 * </ul>
 *
 * @version	$Rev: 76 $ $Date: 2006-01-17 18:15:44 +0100 (mar., 17 janv. 2006) $
 * @since	Petals 1.0
 * @author	<a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public interface JBIPackage {

	/**
	 * Returns the URI indicating the location to package Zip archive.
	 *
	 * @return <code>URI</code> the package Zip archive location.
	 */
	public URI getArchiveURI();

	/**
	 * Returns <code>true</code> if this is an installation package ZIP archive.
	 * The installation package can be either for a JBI component or a
	 * shared-library.
	 *
	 * @return <code>true</code> if this is a installation package. Otherwise,
	 * returns <code>false</code>.
	 */
	public boolean isInstallationPackage();

	/**
	 * Returns <code>true</code> if this is a deployment package ZIP archive.
	 * Such package is used to bundle a service assembly deployment and its
	 * service units entries.
	 *
	 * @return <code>true</code> if this is a deployment package. Otherwise,
	 * returns <code>false</code>.
	 */
	public boolean isDeploymentPackage();

	/**
	 * Get the JBI Descriptor entry for this package. It's a populated object
	 * instance after loading, parsing and validating the descriptor located in
	 * <code>/META-INF/jbi.xml</code>.
	 *
	 * @return <code>JBIDescriptor</code> loaded descriptor.
	 * @throws PetalsException in case an error occur while loading, parsing
	 * or validating the given package's JBI Descriptor against its schema.
	 */
	public JBIDescriptor getDescriptor();

	/**
	 * The method checkContents() provided by this interface to improve the
	 * package validation. Invoking this method ensure that the JBI Package is
	 * consistent to the informations provided in the descriptor.
	 * <p>
	 * The general idea is checking if specified file entries specified in
	 * the descriptor exists within the ZIP archive. Additionally, each
	 * implementation can include other validation checkings according to the
	 * JBI specification.
	 *
	 * @throws PetalsException
	 */
	public void checkContents() throws PetalsException;

}
