/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005  EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: DeliveryChannelImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.messaging;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * TODO DeliveryChannelImpl:check multi thread use
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 */
public class DeliveryChannelImpl implements DeliveryChannel {
	
	/**
	 * Monolog Logger instance.
	 */
	protected Logger logger;
	
	protected LoggingUtil log;
	
	protected Router router;
	
	protected IncomingMessageQueue queue;
	
	protected ComponentContextImpl context;
	
	protected boolean opened = true;
	
	/**
	 * This messageExchangeFactory is used
	 * to return <code>MessageExchangeDecorator</code> on
	 * accept() methods.
	 */
	protected MessageExchangeFactoryImpl messageExchangeFactory;
	
	/**
	 * Constructor
	 * @param c
	 */
	public DeliveryChannelImpl(ComponentContextImpl context,Router router,Logger logger) 
	{
		this.logger = logger;
		
		log = new LoggingUtil(logger);
		
		queue = new IncomingMessageQueue(logger);
		
		messageExchangeFactory = new MessageExchangeFactoryImpl(this,logger);
		
		this.router = router;
		this.context = context;
	}
	
	/**
	 * return a MessageExchangeDecorator with 'consumer' or 'provider'
	 * observer role (depending on the messageExchange getRole())
	 * @see javax.jbi.messaging.DeliveryChannel#accept()
	 */
	public MessageExchange accept() throws MessagingException 
	{
		log.call();
		
		return accept(0);
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#accept(long)
	 * TODO accept:synchronize
	 */
	public MessageExchange accept(long timeoutMS) throws MessagingException 
	{
		log.start();
		
		checkDeliveryChannelIsOpened();
		
		MessageExchangeImpl msg;
		
		try 
		{
			msg = queue.poll(timeoutMS);
		} 
		catch (Exception e) 
		{
			throw new MessagingException(e);
		}
		log.end();
		
		return messageExchangeFactory.createExchangeDecorator(msg,msg.getPattern());
	}
	
	/**
	 * (non-Javadoc)
	 * @see javax.jbi.messaging.DeliveryChannel#close()
	 * TODO close:manage the received messages which are not consummed and are still in the incoming message queue.
	 */
	public synchronized void close() throws MessagingException 
	{
		log.call();
		
		opened = false;
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory()
	 */
	public MessageExchangeFactory createExchangeFactory() 
	{
		log.call();
		
		MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,logger);
		
		return mef;
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.xml.namespace.QName)
	 */
	public MessageExchangeFactory createExchangeFactory(QName interfaceName) 
	{
		log.call(interfaceName);
		
		MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,logger);
		
		mef.setDefaultInterfaceName(interfaceName);

		return mef;
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.jbi.servicedesc.ServiceEndpoint)
	 */
	public MessageExchangeFactory createExchangeFactory(ServiceEndpoint endpoint) 
	{
		log.call(endpoint);
		
		MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,logger);
		
		mef.setDefaultServiceEndpoint(endpoint);
		
		return mef;
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactoryForService(javax.xml.namespace.QName)
	 */
	public MessageExchangeFactory createExchangeFactoryForService(
			QName serviceName) 
	{
		log.call(serviceName);
		
		MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,logger);
		
		mef.setDefaultServiceName(serviceName);
		
		return mef;
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#send(javax.jbi.messaging.MessageExchange)
	 */
	public void send(MessageExchange exchange) throws MessagingException 
	{
		log.start(exchange);
		
		checkDeliveryChannelIsOpened();
		checkMessageExchange(exchange);
		
		MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;
		MessageExchangeImpl exchangeImpl = exchangeDecorator.getMessageExchange();
		
		try 
		{
			router.send(context,exchangeImpl);
			// now the message has been sent,
			// check if the the exchange has to become terminated
			if(! ExchangeStatus.ACTIVE.equals(exchangeImpl.getStatus()))
			{
				exchangeImpl.setTerminated(true);
			}
		}
		catch (JBIException e) 
		{
			throw new MessagingException(e);
		}
		log.end();
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#sendSynch(javax.jbi.messaging.MessageExchange)
	 */
	public boolean sendSync(MessageExchange exchange) throws MessagingException 
	{
		log.call();
		return sendSync(exchange,0);
	}
	
	/**
	 * @see javax.jbi.messaging.DeliveryChannel#sendSynch(javax.jbi.messaging.MessageExchange, long)
	 */
	public boolean sendSync(MessageExchange exchange, long timeoutMS)
	throws MessagingException 
	{
		log.start();
	
		MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;
		
		synchronized (exchangeDecorator) 
		{
			try 
			{
				// Prepare the send
				exchangeDecorator.setWaitingOnSynchronousSend(true);
				send(exchange);
				
				// wait for response or timeout
				exchangeDecorator.wait(timeoutMS);
				
				// Now, the response has been sent (wait unblocked),
				// or timeout occured
				
				if(exchangeDecorator.isWaitingOnSynchronousSend())
				{
					// we were always waiting for the response
					
					// get the ownership of the ME
					exchangeDecorator.getMessageExchange().setRole(exchangeDecorator.getObserverRole());
					
					exchangeDecorator.setStatus(ExchangeStatus.ERROR);
					
					exchangeDecorator.getMessageExchange().setTerminated(true);
				}
				
			} catch (InterruptedException e) 
			{
				log.error(e);
				// get the ownership of the ME
				exchangeDecorator.getMessageExchange().setRole(exchangeDecorator.getObserverRole());
				
				exchangeDecorator.setStatus(ExchangeStatus.ERROR);
				
				exchangeDecorator.getMessageExchange().setTerminated(true);
				
				throw new MessagingException(e);
			}
		}
		log.end();
		
		return (! exchangeDecorator.isWaitingOnSynchronousSend());
	}
	
	
	
	/////////////
	// non jbi
	////////////
	
	public synchronized boolean isOpened() {
		return opened;
	}
	
	/**
	 * Check if the <code>DeliveryChannel</code> is opened
	 * @throws MessagingException
	 */
	public synchronized void checkDeliveryChannelIsOpened()
	throws MessagingException {
		if (!opened) {
			throw new MessagingException("DeliveryChannel is closed.");
		}
	}
	
	/**
	 * Check if the <code>MessageExchange</code> is Ok
	 * (not null and not terminated)
	 * @throws MessagingException
	 */
	public void checkMessageExchange(MessageExchange exchange)
	throws MessagingException 
	{
		if (exchange == null) 
		{
			throw new MessagingException("MessageExchange is null.");
		}
		MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;
		
		if(exchangeDecorator.getMessageExchange().isTerminated())
		{
			throw new MessagingException("MessageExchange is terminated.");
		}
	}
	
	/**
	 * Used by the <code>Transporter</code> when a
	 * message is received by the <code>Transporter</code> and the
	 * recepient of the message is the component attached to this channel.
	 * The message is put in the channel queue, and will be consummed
	 * when the component calls an <i>accept</i> mehod;
	 * or if its a response of a sendSync() method, unblock it.
	 * @param msg the message received by the <code>Transporter</code>
	 * @throws JBIException the channel has been closed by the component. The message can not
	 * be accepted
	 */
	public void push(MessageExchangeImpl msg) throws JBIException 
	{
		log.start(msg);
		
		checkDeliveryChannelIsOpened();
		
		// check if this ME is in a synchronized state
		// so we must unblock the sendSync() method,

		// or if this ME is asynchronized,
		// and we just do the accept().
		
		if(Role.PROVIDER.equals(msg.getRole()))
		{
			// we are in a "provider" context, so
			// check if the ME has a "ProviderMEDecorator"
			// which was set previously,
			// and which is waiting for this response.
			MessageExchangeDecorator providerMEDecorator = msg.getProviderMEDecorator();

			if(providerMEDecorator!= null
					&& providerMEDecorator.isWaitingOnSynchronousSend())
			{
				// we have to notify the ProviderMEDecorator
				// that its response is received.
				synchronized (providerMEDecorator) 
				{
					providerMEDecorator.setWaitingOnSynchronousSend(false);
					providerMEDecorator.notifyAll();
				}
			}
			else
			{
				// the providerMEDecorator is null 
				// or was not waiting in a synchronously way
				queue.push(msg);
			}
		}
		
		else if(Role.CONSUMER.equals(msg.getRole()))
		{
			// we are in a "consumer" context, so
			// check if the ME has a "ConsumerMEDecorator"
			// which was set previously,
			// and which is waiting for this response.
			MessageExchangeDecorator consumerMEDecorator = msg.getConsumerMEDecorator();
			
			if(consumerMEDecorator!= null
					&& consumerMEDecorator.isWaitingOnSynchronousSend())
			{
				// we have to notify the ProviderMEDecorator
				// that its response is received.
				synchronized (consumerMEDecorator)
				{
					consumerMEDecorator.setWaitingOnSynchronousSend(false);
					consumerMEDecorator.notifyAll();
				}
			}
			else
			{
				// the consumerMEDecorator is null 
				// or was not waiting in a synchronously way
				queue.push(msg);
			}
		}
		log.end();
	}
}
