/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005  EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AddressResolverImpl.java,v 1.1 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;
import javax.jbi.component.Component;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.registry.RegistryManager;

/**
 * The <code>AddressResolver</code> resolves implicit addresses such
 * as <code>interface</code> or <code>service</code> into an <code>endpoint</code>.
 * Rules are used to help the <code>AddressResolver</code> in the choice of
 * the <code>endpoint</code>, if more than one implement the implicit address.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 *
 */
public class AddressResolver 
{
	
	protected RegistryManager registry;
	
	
	public AddressResolver(RegistryManager reg)
	{
		registry=reg;
	}
	
	/**
	 * Resolve the address if necessary.
	 * Resolution consist in finding an endpoint satisfying
	 * <code>interface</code> or <code>service</code>.
	 * the source of the exchange and the exchange help
	 * the <code>AddressResolver</code> to find the endpoint.
	 * The <code>containerName</code> and <code>ComponentName</code>
	 * properties are set.
	 * If <code>containerName</code> and <code>ComponentName</code> 
	 * already exist, do nothing.
	 * The check of acceptance from the two part is done here
	 * (the Component method isExchangeWithXXXOkay()). 
	 * @param address
	 * @param source
	 * @param exchange
	 */
	public void resolveAddress(ComponentContextImpl source, MessageExchangeImpl exchange) throws RoutingException
	{
		ServiceEndpoint endpoint 	= exchange.getEndpoint();
		QName serviceName 			= exchange.getService();
		QName interfaceName 		= exchange.getInterfaceName();
		
		// Construct an array of potential matching endpoints.
		// If the endpoint has been set in the messageExchange,
		// this array contains only one element, which is this endpoint.
		ServiceEndpoint[] endpoints = null;
		
		if (endpoint == null)
		{
			if (serviceName != null)
			{
				// find an Endpoint with the specified ServiceName
				endpoints = registry.getInternalEndpointsForService(serviceName);
				
				if (endpoints == null || endpoints.length == 0)
				{     
					throw new RoutingException("No endpoint found for the specified service");
				}
			}
			else if (interfaceName != null)
			{                
				// find an Endpoint with the specified InterfaceName
				endpoints = registry.getInternalEndpointsForInterface(interfaceName);
				
				if (endpoints.length == 0)
				{
					throw new RoutingException("No endpoint found for the specified interface");
				}
			}
			else
			{      
				throw new RoutingException("No information about destination is specified");
			}
		}
		else
		{
			// The endpoint is specified by the Component.
			// a Jbi 'connection' might has been defined for this endpoint,
			// so try to resolve the real address
			ServiceEndpoint linkedResolvedEndpoint = resolveLinkedAddress(source,endpoint);
			
			if(linkedResolvedEndpoint != null)
			{
				endpoints = new ServiceEndpoint[]{linkedResolvedEndpoint};
			}
			else
			{
				// there was no Jbi 'connection' defined,
				// so this endpoint is a real address.
				
				// The endpoint is specified by the Component, but it might have been created by this component,
				// so it might not be a Petals implementation of 'ServiceEndpoint'. 
				// We have to get the Endpoint that matching it, from the Registry.
				ServiceEndpoint ep = registry.getInternalEndpoint(endpoint.getServiceName(),endpoint.getEndpointName()); 
				
				if (ep != null)
				{
					endpoints = new ServiceEndpoint[]{ep};
				}
				else
				{
					throw new RoutingException("The specified Endpoint does not match a registered endpoint.");
				}
			}
			
		}
		// if there is several endpoints that matching
		// the given information, we have to choose one.
		endpoint = electEndpoint(endpoints, source, exchange);
		
		if(endpoint == null)
		{
			throw new RoutingException("Consumer and/or Provider refused the Exchange.");
		}
		// Now, we can set the real endpoint that will
		// serves the MessageExchange
		exchange.setEndpoint(endpoint);
	}
	
	/**
	 * If the specified endpoint corresponds to a logical address
	 * (jbi connection), return the real address.
	 * Otherwise, return null.
	 * TODO : resolve "connections"
	 * @param source
	 * @param exchange
	 */
	private ServiceEndpoint resolveLinkedAddress(ComponentContextImpl source, ServiceEndpoint endpoint)
	{
		return null;
	}
	
	/**
	 * Elects the address in the list of endpoints as the message recipient, with the
	 * defined strategy.
	 * Ask the consumer and the provider if they are Okay for
	 * the exchange.
	 * TODO electAddress: for the moment, return the first element satisfying the "isExchangeWithXXXOkay"
	 * @param addresses
	 * @return the elected recipient 
	 */
	private ServiceEndpoint electEndpoint(ServiceEndpoint[] endpoints,ComponentContextImpl ctxSource,MessageExchange exchange)
	{
		ServiceEndpoint result = null;
		
		if(endpoints != null && endpoints.length>0)
		{
			Component consumer = ctxSource.getComponent();
			
			for (int i = 0; i < endpoints.length && result == null; i++) 
			{
				InternalEndpoint providerEP = (InternalEndpoint)endpoints[i];
				
				Component provider = providerEP.getComponentContext().getComponent();
				
				if (consumer.isExchangeWithProviderOkay(providerEP,exchange)
				 && provider.isExchangeWithConsumerOkay(providerEP,exchange))
				{
					result = providerEP;
				}
			}
		}
		return result;
	}
}
