/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005  EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: RouterImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;


import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_REGISTRY_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_TRANSPORTER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.jbi.component.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.RegistryManager;
import org.objectweb.petals.jbi.transport.TransportException;
import org.objectweb.petals.jbi.transport.Transporter;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;



/**
 * Routes messages to their destinations.
 * Performs an itinerary resolution, adding for example a transformatino service
 * before delivering the message to its recipient,
 * and perform an address resolution, finding the endpoint satisfying the
 * the destination information (finding an endpoint for a specified service).
 * Resolutions are based on defined rules.
 * @author Adrien LOUIS - EBM WebSourcing
 */
public class RouterImpl implements Router, BindingController, LifeCycleController
{
	private static final String[] CLIENT_INTERFACES = new String[] {
		JBI_REGISTRY_NAME };

	
	private Logger logger;
	
	private RegistryManager registry;
	
    private AddressResolver addressResolver=null;
        
    private Transporter transporter = null;
    
    /**
	 * Component LifeCycle state.
	 */
	private String fcState;
	
	
    /**
     * Sends the specified message.
     * Depending on the Role of the sender,
     * send the message to the producer (request)
     * or to the consumer (response)
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange) throws RoutingException
    {
    	LoggingUtil.call(logger);
    	
    	if(MessageExchange.Role.CONSUMER.equals(exchange.getRole()))
    	{
    		sendToProvider(source,exchange);
    	}
    	else
    	{
    		sendToConsumer(source,exchange);
    	}
     }   
    
    
    /**
     * Sends the specified message to the provider of the service.
     * TODO for the moment, send directly to the DeliveryChannel of the component related to the recipient endpoint.
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    private void sendToProvider(ComponentContextImpl source, MessageExchangeImpl exchange) throws RoutingException
    {
    	LoggingUtil.start(logger,source+", "+exchange);
    	
    	// resolve the address (find endpoint)
    	addressResolver.resolveAddress(source,exchange);
    	
    	try {
			transporter.sendToProvider(exchange);
		} catch (TransportException e) {
			throw new RoutingException(e);
		}
    	
		LoggingUtil.end(logger);
     }   
    
    
    /**
     * Sends the specified message, sent by the specified component.
     * TODO send directly to the DeliveryChannel of the component related to the recipient endpoint.
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    private void sendToConsumer(ComponentContextImpl source, MessageExchangeImpl exchange) throws RoutingException
    {
    	LoggingUtil.start(logger,source+", "+exchange);
    	
    	try 
    	{
			transporter.sendToConsumer(exchange);
		} 
    	catch (TransportException e) 
    	{
			throw new RoutingException(e);
		}
    	
		LoggingUtil.end(logger);
     }   
    
    
    
    
    
    
    
    //  ----------------------------------------------------------------------
	//   Fractal BindingController implementation
	// ----------------------------------------------------------------------


	/**
	 * @see org.objectweb.fractal.api.control.BindingController#listFc()
	 */
	public String[] listFc() {
		return CLIENT_INTERFACES;
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 */
	public Object lookupFc(String clientItfName)
			throws NoSuchInterfaceException {

		if (JBI_REGISTRY_NAME.equals(clientItfName)){
			return registry;
		}
		else if (JBI_TRANSPORTER_NAME.equals(clientItfName)){
				return transporter;
				
		}else{
			throw new NoSuchInterfaceException(clientItfName);
		}	
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String, java.lang.Object)
	 */
	public void bindFc(String clientItfName, Object serverItf)
			throws NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException {

		if (LOGGER_NAME.equals(clientItfName)) {
			logger = (Logger) serverItf;
		}else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
			//nothing
		}else if (JBI_REGISTRY_NAME.equals(clientItfName)) {
			registry = (RegistryManager) serverItf;
		}else if (JBI_TRANSPORTER_NAME.equals(clientItfName)) {
			transporter = (Transporter) serverItf;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	/**
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {

		if (JBI_REGISTRY_NAME.equals(clientItfName)) {
			registry = null;
		}
		else if (JBI_TRANSPORTER_NAME.equals(clientItfName)) {
			transporter = null;
		}
		else {
			throw new NoSuchInterfaceException(clientItfName);
		}
	}

	// ----------------------------------------------------------------------
	//   Fractal LifecycleController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return (STARTED.equals(fcState) ? STARTED : STOPPED);
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {

    	addressResolver = new AddressResolver(registry);

		fcState = STARTED;
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {

		fcState = STOPPED;
	}
}
