/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsClassLoader.java 216 2006-04-13 14:18:14Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.classloader;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.objectweb.petals.classloader.factory.URLFactory;
import org.objectweb.petals.classloader.locator.Locator;

/**
 * @version	$Rev: 216 $ $Date: 2006-04-13 16:18:14 +0200 (jeu., 13 avr. 2006) $
 * @since	Petals 1.0
 * @author	Rafael Marins, Olivier Fabre
 */
public class PetalsClassLoader extends URLClassLoader {

	/**
	 * Standard Java package binary name.
	 */
	private static final String JAVA_PACKAGE = "java.";

	/**
	 * Standard Java extensions package binary name.
	 */
	private static final String JAVAX_PACKAGE = "javax.";

	/**
	 * To use or not the parent first class loading delegation model.
	 */
	private boolean parentFirst;

	/**
	 * Isolate Class Loading from parent.
	 */
	private boolean ignoreParent;

	/**
	 * To use or not the package intelligence self-first mechanism.
	 */
	private boolean usePackageIntelligence;

	/**
	 * List of known packages loaded from this class loader.
	 */
	private Set<String> localKnownPackages;

	/**
	 * Parent Class Loader.
	 */
	private ClassLoader parent;
	
    /** 
     * Factories used to create URLs for URLClassLoader 
     */
    private URLFactory[] factories;

    /** 
     * Locators used to search jar/dir for files, directories, ... 
     */
    private Locator[] locators;

    /** 
     * Bases URLs 
     */
    private URL[] bases;

	/**
	 * 
	 * @param baseUrls
	 * @param classpathLocations
	 * @param parent
	 * @throws IOException
	 */
	public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations, ClassLoader parent) throws IOException {
		super(new URL[0], parent);
		this.parent = parent;
		this.bases = baseUrls;
		initialize(classpathLocations);
	}

	/**
	 * 
	 * @param baseUrls
	 * @param classpathLocations
	 * @throws IOException
	 */
	public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations) throws IOException {
		super(new URL[0]);
		this.bases = baseUrls;
		initialize(classpathLocations);
	}

	/**
	 * 
	 * @param baseUrls
	 * @param classpathLocations
	 * @param parent
	 * @param factory
	 * @throws IOException
	 */
	public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations, ClassLoader parent,
			URLStreamHandlerFactory factory) throws IOException {
		super(new URL[0], parent, factory);
		this.parent = parent;
		this.bases = baseUrls;
		initialize(classpathLocations);
	}

	// ----------------------------------------------------------------------
	//   Getters and Setters
	// ----------------------------------------------------------------------

	/**
	 * @return Returns the useParentFirst.
	 */
	public boolean isParentFirst() {
		return parentFirst;
	}

	/**
	 * @param use The useParentFirst to set.
	 */
	public void setParentFirst(boolean use) {
		this.parentFirst = use;
	}

	/**
	 * @return Returns the isolated.
	 */
	public boolean isIsolated() {
		return ignoreParent;
	}

	/**
	 * @param isolated The ignoreParent to set.
	 */
	public void setIsolated(boolean isolated) {
		this.ignoreParent = isolated;
	}

	/**
	 * @return Returns the usePackageIntelligence.
	 */
	public boolean isUsingPackageIntelligence() {
		localKnownPackages = null;
		return usePackageIntelligence;
	}

	/**
	 * @param use The usePackageIntelligence to set.
	 */
	public void setUsingPackageIntelligence(boolean use) {
		localKnownPackages = new HashSet<String>();
		this.usePackageIntelligence = use;
	}
	
    /**
     * @return Returns the bases.
     */
    public URL[] getBases() {
        return bases;
    }
    
    /**
     * @return Returns a String representation of the classpath used by this classloader
     */
    public String getClasspath() {
        URL[] urls = getURLs();
        StringBuffer cp = new StringBuffer();
        for (int i = 0; i < urls.length; i++) {
            String url = urls[i].getFile();
            // do not add URL with !/ inside
            if (url.indexOf("!/") == -1) {
                cp.append(File.pathSeparator + url);
            }
        }
        return  cp.toString();
    }

	// ----------------------------------------------------------------------
	//   overriding methods
	// ----------------------------------------------------------------------

	/**
	 * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
	 */
	@Override
	protected synchronized Class<?> loadClass(String className, boolean resolve)
			throws ClassNotFoundException {

		Class<?> clazz = null;

		clazz = findLoadedClass(className);

		if (clazz == null) {
			// parent first
			if (useParentFirst(className, true)) {
				try {
					clazz = parent.loadClass(className);
				}
				catch (ClassNotFoundException cnfe) {
					clazz = findClass(className);
				}
			}
			// self first
			else {
				try {
					clazz = findClass(className);
				}
				catch (ClassNotFoundException cnfe) {
					if (isIsolated()) {
						throw cnfe;
					}
					else {
						clazz = parent.loadClass(className);
					}
				}
			}
		}

		if (resolve) {
			resolveClass(clazz);
		}

		return clazz;
	}

	/**
	 * @see java.net.URLClassLoader#findClass(java.lang.String)
	 */
	@Override
	protected Class<?> findClass(String className) throws ClassNotFoundException {

		String packageName = getPackageName(className);
		addLocalKnownPackage(packageName);

		return super.findClass(className);
	}

	/**
	 * @see java.lang.ClassLoader#getResource(java.lang.String)
	 */
	@Override
	public URL getResource(String resourceName) {

		URL resource = null;

		// parent first
		if (useParentFirst(resourceName, false)) {
			resource = parent.getResource(resourceName);

			if (resource == null) {
				resource = findResource(resourceName);
			}
		}
		// self first
		else {
			resource = findResource(resourceName);

			if (resource == null && !isIsolated()) {
				resource = parent.getResource(resourceName);
			}
		}

		return resource;
	}

	/**
	 * @see java.lang.ClassLoader#getResources(java.lang.String)
	 */
	@Override
	public Enumeration<URL> getResources(String resourceName) throws IOException {

		Enumeration<URL> resources = null;

		// parent first
		if (useParentFirst(resourceName, false)) {
			resources = parent.getResources(resourceName);

			if (resources == null) {
				resources = findResources(resourceName);
			}
		}
		// self first
		else {
			resources = findResources(resourceName);

			if (resources == null && !isIsolated()) {
				resources = parent.getResources(resourceName);
			}
		}

		return resources;
	}

	// ----------------------------------------------------------------------
	//   implementation specific methods
	// ----------------------------------------------------------------------

	private void initialize(List<String> classpathLocations) throws IOException {
		this.parentFirst = true;
		this.ignoreParent = false;
		this.usePackageIntelligence = false;

		if (parent == null) {
			parent = getSystemClassLoader();
		}
		
		factories = new URLFactory[bases.length];
        locators = new Locator[bases.length];

        // Create factories and locator for each base URL
        for (int i = 0; i < bases.length; i++) {

            factories[i] = URLFactory.getFactory(bases[i]);
            locators[i] = Locator.getLocator(bases[i]);
            
        }
        
        // Add all classpathLocations into the repository
        for (String location : classpathLocations) {

        	addInRepository(location);
        	
        }
	}

	private boolean useParentFirst(String resourceName, boolean usingForClass) {

		// begin with the chosen delegation model
        boolean useParentFirst = isParentFirst() && !isIsolated();

        // try self-first based on intelligent mechanism of known packages 
        if (useParentFirst) {

        	if (isUsingPackageIntelligence() && usingForClass) {

        		String packageName = getPackageName(resourceName);

        		// self-first <-- when parent first is set to false
        		useParentFirst = !knowLocalPackage(packageName);
        	}
        }
        // always use parent class loader when loading java packages
        else {
        	useParentFirst = (JAVA_PACKAGE.startsWith(resourceName) ||
        			JAVAX_PACKAGE.startsWith(resourceName));
        }

        return useParentFirst;
	}

	private void addLocalKnownPackage(String packageName) {
		if (localKnownPackages != null) {
			localKnownPackages.add(packageName);
		}
	}

	private boolean knowLocalPackage(String packageName) {

		boolean known = false;

		if (localKnownPackages != null) {
			known = localKnownPackages.contains(packageName);
		}

		return known;
	}

	private String getPackageName(String className) {

		int lastSeparator = className.lastIndexOf('.');
		String packageName = className.substring(0, lastSeparator);

		return packageName;
	}
	
    /**
     * Add specified location into the repository.
     * If location is found in multiple URLs of the bases, new
     * URLs will be added multiple times.
     *
     * @param location an entry name (for a jar) or a path name (for a directory)
     *
     * @throws IOException when constructed URL is malformed
     */
    protected void addInRepository(String location) throws IOException {
        // test existence in each base URL
        for (int i = 0; i < bases.length; i++) {
            if (locators[i].hasDirectory(location)
                || locators[i].hasFile(location)) {
                addURL(factories[i].getURL(location));
            }
        }
    }

    /**
     * Add specified location childs into the repository.
     * Search will be performed on the bases URLs.
     *
     * @param location an entry name (for a jar) or a path name (for a directory)
     *
     * @throws IOException When a directory content cannot be explored.
     *
     * TODO test existence is now done in AbsModuleClassLoader
     */
    protected void addContentInRepository(String location) throws IOException {
        // test existence in each base URL
        for (int i = 0; i < bases.length; i++) {
            List list = locators[i].listContent(location);
            for (Iterator l = list.iterator(); l.hasNext();) {
                addURL(factories[i].getURL((String) l.next()));
            }
        }
    }

    /**
     * @return Returns a String representation of the AbsModuleClassLoader
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append("classloader : " + getClass().getName() + "\n");
        sb.append("\tmodules bases (not in loader!) : \n");
        for (int i = 0; i < bases.length; i++) {
            sb.append("\t\t -" + bases[i] + "\n");
        }
        sb.append("\trepositories :\n");
        URL[] rep = getURLs();
        for (int i = 0; i < rep.length; i++) {
            sb.append("\t\t -" + rep[i] + "\n");
        }
        sb.append("\tparent : " + getParent() + "\n");

        return sb.toString();
    }

}
